"""
>>> from webscout import YTTranscriber
>>> transcript = YTTranscriber.get_transcript('https://www.youtube.com/watch?v=dQw4w9WgXcQ')
>>> print(transcript)
{'text': 'Never gonna give you up', 'start': 0.0, 'duration': 4.5}

"""

import html
import http.cookiejar as cookiejar
import re
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from typing import Dict, List, Optional, Union
from xml.etree import ElementTree

from curl_cffi.requests import Session

from webscout.exceptions import (
    CookiePathInvalidError,
    FailedToCreateConsentCookieError,
    InvalidVideoIdError,
    NoTranscriptFoundError,
    NotTranslatableError,
    TooManyRequestsError,
    TranscriptRetrievalError,
    TranscriptsDisabledError,
    TranslationLanguageNotAvailableError,
    VideoUnavailableError,
    YouTubeRequestFailedError,
)
from webscout.litagent import LitAgent

# YouTube API settings
WATCH_URL = 'https://www.youtube.com/watch?v={video_id}'
INNERTUBE_API_URL = "https://www.youtube.com/youtubei/v1/player?key={api_key}"
INNERTUBE_CONTEXT = {"client": {"clientName": "ANDROID", "clientVersion": "20.10.38"}}
MAX_WORKERS = 4


class YTTranscriber:
    """Transcribe YouTube videos with style! 🎤

    >>> transcript = YTTranscriber.get_transcript('https://youtu.be/dQw4w9WgXcQ')
    >>> print(transcript[0]['text'])
    'Never gonna give you up'
    """

    _session = None
    _executor = ThreadPoolExecutor(max_workers=MAX_WORKERS)

    @classmethod
    def _get_session(cls):
        if cls._session is None:
            cls._session = Session()
            cls._session.headers.update({
                'User-Agent': LitAgent().random()
            })
        return cls._session

    @classmethod
    @lru_cache(maxsize=100)
    def get_transcript(cls, video_url: str, languages: Optional[str] = 'en',
                      proxies: Optional[Dict[str, str]] = None,
                      cookies: Optional[str] = None,
                      preserve_formatting: bool = False) -> List[Dict[str, Union[str, float]]]:
        """
        Retrieves the transcript for a given YouTube video URL.

        Args:
            video_url (str): YouTube video URL (supports various formats).
            languages (str, optional): Language code for the transcript.
                                    If None, fetches the first available transcript.
                                    Defaults to 'en'.
            proxies (Dict[str, str], optional): Proxies to use for the request. Defaults to None.
            cookies (str, optional): Path to the cookie file. Defaults to None.
            preserve_formatting (bool, optional): Whether to preserve formatting tags. Defaults to False.

        Returns:
            List[Dict[str, Union[str, float]]]: A list of dictionaries, each containing:
                - 'text': The transcribed text.
                - 'start': The start time of the text segment (in seconds).
                - 'duration': The duration of the text segment (in seconds).

        Raises:
            TranscriptRetrievalError: If there's an error retrieving the transcript.
        """
        video_id = cls._extract_video_id(video_url)
        http_client = cls._get_session()

        if proxies:
            http_client.proxies.update(proxies)

        if cookies:
            cls._load_cookies(cookies, video_id)

        transcript_list = TranscriptListFetcher(http_client).fetch(video_id)
        language_codes = [languages] if languages else None
        transcript = transcript_list.find_transcript(language_codes)

        return transcript.fetch(preserve_formatting)

    @staticmethod
    def _extract_video_id(video_url: str) -> str:
        """Extracts the video ID from different YouTube URL formats."""
        patterns = [
            r'(?:v=|\/)([0-9A-Za-z_-]{11}).*',
            r'youtu\.be\/([0-9A-Za-z_-]{11})',
            r'youtube\.com\/embed\/([0-9A-Za-z_-]{11})',
            r'youtube\.com\/shorts\/([0-9A-Za-z_-]{11})'
        ]

        for pattern in patterns:
            match = re.search(pattern, video_url)
            if match:
                return match.group(1)

        if re.match(r'^[0-9A-Za-z_-]{11}$', video_url):
            return video_url

        raise InvalidVideoIdError(video_url)

    @staticmethod
    def _load_cookies(cookies: str, video_id: str) -> cookiejar.MozillaCookieJar:
        """Loads cookies from a file."""
        try:
            cj = cookiejar.MozillaCookieJar(cookies)
            cj.load()
            return cj
        except (cookiejar.LoadError, FileNotFoundError):
            raise CookiePathInvalidError(video_id)


class TranscriptListFetcher:
    """Fetches the list of transcripts for a YouTube video using InnerTube API."""

    def __init__(self, http_client: Session):
        """Initializes TranscriptListFetcher."""
        self._http_client = http_client

    def fetch(self, video_id: str):
        """Fetches and returns a TranscriptList."""
        captions_json = self._fetch_captions_json(video_id)
        return TranscriptList.build(
            self._http_client,
            video_id,
            captions_json,
        )

    def _fetch_captions_json(self, video_id: str) -> dict:
        """Fetches captions JSON using InnerTube API."""
        # First get the HTML to extract the API key
        video_html = self._fetch_video_html(video_id)
        api_key = self._extract_innertube_api_key(video_html, video_id)

        # Use InnerTube API to get video data
        innertube_data = self._fetch_innertube_data(video_id, api_key)
        return self._extract_captions_from_innertube(innertube_data, video_id)

    def _extract_innertube_api_key(self, html_content: str, video_id: str) -> str:
        """Extracts the InnerTube API key from HTML."""
        pattern = r'"INNERTUBE_API_KEY":\s*"([a-zA-Z0-9_-]+)"'
        match = re.search(pattern, html_content)
        if match and len(match.groups()) == 1:
            return match.group(1)

        # Check for IP block
        if 'class="g-recaptcha"' in html_content:
            raise TooManyRequestsError(video_id)

        raise TranscriptRetrievalError(video_id, "Could not extract InnerTube API key")

    def _fetch_innertube_data(self, video_id: str, api_key: str) -> dict:
        """Fetches video data from InnerTube API."""
        response = self._http_client.post(
            INNERTUBE_API_URL.format(api_key=api_key),
            json={
                "context": INNERTUBE_CONTEXT,
                "videoId": video_id,
            },
        )
        return _raise_http_errors(response, video_id).json()

    def _extract_captions_from_innertube(self, innertube_data: dict, video_id: str) -> dict:
        """Extracts captions JSON from InnerTube API response."""
        # Check playability status
        playability_status = innertube_data.get("playabilityStatus", {})
        status = playability_status.get("status")

        if status == "ERROR":
            reason = playability_status.get("reason", "Unknown error")
            if "unavailable" in reason.lower():
                raise VideoUnavailableError(video_id)
            raise TranscriptRetrievalError(video_id, reason)

        if status == "LOGIN_REQUIRED":
            reason = playability_status.get("reason", "")
            if "bot" in reason.lower():
                raise TooManyRequestsError(video_id)
            if "age" in reason.lower() or "inappropriate" in reason.lower():
                raise TranscriptRetrievalError(video_id, "Video is age-restricted")
            raise TranscriptRetrievalError(video_id, reason or "Login required")

        # Get captions
        captions = innertube_data.get("captions", {})
        captions_json = captions.get("playerCaptionsTracklistRenderer")

        if captions_json is None or "captionTracks" not in captions_json:
            raise TranscriptsDisabledError(video_id)

        return captions_json

    def _create_consent_cookie(self, html_content, video_id):
        match = re.search('name="v" value="(.*?)"', html_content)
        if match is None:
            raise FailedToCreateConsentCookieError(video_id)
        self._http_client.cookies.set('CONSENT', 'YES+' + match.group(1), domain='.youtube.com')

    def _fetch_video_html(self, video_id):
        html_content = self._fetch_html(video_id)
        if 'action="https://consent.youtube.com/s"' in html_content:
            self._create_consent_cookie(html_content, video_id)
            html_content = self._fetch_html(video_id)
            if 'action="https://consent.youtube.com/s"' in html_content:
                raise FailedToCreateConsentCookieError(video_id)
        return html_content

    def _fetch_html(self, video_id):
        response = self._http_client.get(WATCH_URL.format(video_id=video_id), headers={'Accept-Language': 'en-US'})
        return html.unescape(_raise_http_errors(response, video_id).text)


class TranscriptList:
    """
    >>> transcript_list = TranscriptList.build(http_client, video_id, captions_json)
    >>> transcript = transcript_list.find_transcript(['en'])
    >>> print(transcript)
    en ("English")[TRANSLATABLE]
    """

    def __init__(self, video_id, manually_created_transcripts, generated_transcripts, translation_languages):
        """Init that transcript list with all the good stuff! 💯"""
        self.video_id = video_id
        self._manually_created_transcripts = manually_created_transcripts
        self._generated_transcripts = generated_transcripts
        self._translation_languages = translation_languages

    @staticmethod
    def build(http_client, video_id, captions_json):
        """
        Factory method for TranscriptList.

        :param http_client: http client which is used to make the transcript retrieving http calls
        :type http_client: Session
        :param video_id: the id of the video this TranscriptList is for
        :type video_id: str
        :param captions_json: the JSON parsed from the YouTube API
        :type captions_json: dict
        :return: the created TranscriptList
        :rtype TranscriptList:
        """
        # Handle both old format (simpleText) and new format (runs)
        translation_languages = []
        for tl in captions_json.get('translationLanguages', []):
            lang_name = tl.get('languageName', {})
            if isinstance(lang_name, dict):
                # Try new format first (runs), then old format (simpleText)
                if 'runs' in lang_name:
                    name = lang_name['runs'][0]['text']
                elif 'simpleText' in lang_name:
                    name = lang_name['simpleText']
                else:
                    name = tl.get('languageCode', 'Unknown')
            else:
                name = str(lang_name)
            translation_languages.append({
                'language': name,
                'language_code': tl['languageCode'],
            })

        manually_created_transcripts = {}
        generated_transcripts = {}

        for caption in captions_json['captionTracks']:
            if caption.get('kind', '') == 'asr':
                transcript_dict = generated_transcripts
            else:
                transcript_dict = manually_created_transcripts

            # Extract caption name - handle both formats
            caption_name = caption.get('name', {})
            if isinstance(caption_name, dict):
                if 'runs' in caption_name:
                    name = caption_name['runs'][0]['text']
                elif 'simpleText' in caption_name:
                    name = caption_name['simpleText']
                else:
                    name = caption.get('languageCode', 'Unknown')
            else:
                name = str(caption_name) if caption_name else caption.get('languageCode', 'Unknown')

            # Remove &fmt=srv3 from URL as it can cause issues
            base_url = caption['baseUrl'].replace("&fmt=srv3", "")

            transcript_dict[caption['languageCode']] = Transcript(
                http_client,
                video_id,
                base_url,
                name,
                caption['languageCode'],
                caption.get('kind', '') == 'asr',
                translation_languages if caption.get('isTranslatable', False) else [],
            )

        return TranscriptList(
            video_id,
            manually_created_transcripts,
            generated_transcripts,
            translation_languages,
        )

    def __iter__(self):
        return iter(list(self._manually_created_transcripts.values()) + list(self._generated_transcripts.values()))

    def find_transcript(self, language_codes):
        """
        Finds a transcript for a given language code. If no language is provided, it will
        return the first available transcript.

        :param language_codes: A list of language codes in a descending priority.
        :type languages: list[str]
        :return: the found Transcript
        :rtype Transcript:
        :raises: NoTranscriptFound
        """
        if not language_codes:
            language_codes = ['any']

        if 'any' in language_codes:
            for transcript in self:
                return transcript
        return self._find_transcript(language_codes, [self._manually_created_transcripts, self._generated_transcripts])

    def find_generated_transcript(self, language_codes):
        """
        Finds an automatically generated transcript for a given language code.

        :param language_codes: A list of language codes in a descending priority. For example, if this is set to
        ['de', 'en'] it will first try to fetch the german transcript (de) and then fetch the english transcript (en) if
        it fails to do so.
        :type languages: list[str]
        :return: the found Transcript
        :rtype Transcript:
        :raises: NoTranscriptFound
        """
        if not language_codes:
            language_codes = ['any']

        if 'any' in language_codes:
            for transcript in self:
                if transcript.is_generated:
                    return transcript
        return self._find_transcript(language_codes, [self._generated_transcripts])

    def find_manually_created_transcript(self, language_codes):
        """
        Finds a manually created transcript for a given language code.

        :param language_codes: A list of language codes in a descending priority. For example, if this is set to
        ['de', 'en'] it will first try to fetch the german transcript (de) and then fetch the english transcript (en) if
        it fails to do so.
        :type languages: list[str]
        :return: the found Transcript
        :rtype Transcript:
        :raises: NoTranscriptFound
        """
        if not language_codes:
            language_codes = ['any']
        return self._find_transcript(language_codes, [self._manually_created_transcripts])

    def _find_transcript(self, language_codes, transcript_dicts):
        for language_code in language_codes:
            for transcript_dict in transcript_dicts:
                if language_code in transcript_dict:
                    return transcript_dict[language_code]

        raise NoTranscriptFoundError(
            self.video_id,
            language_codes,
            self
        )

    def __str__(self):
        return (
            'For this video ({video_id}) transcripts are available in the following languages:\n\n'
            '(MANUALLY CREATED)\n'
            '{available_manually_created_transcript_languages}\n\n'
            '(GENERATED)\n'
            '{available_generated_transcripts}\n\n'
            '(TRANSLATION LANGUAGES)\n'
            '{available_translation_languages}'
        ).format(
            video_id=self.video_id,
            available_manually_created_transcript_languages=self._get_language_description(
                str(transcript) for transcript in self._manually_created_transcripts.values()
            ),
            available_generated_transcripts=self._get_language_description(
                str(transcript) for transcript in self._generated_transcripts.values()
            ),
            available_translation_languages=self._get_language_description(
                '{language_code} ("{language}")'.format(
                    language=translation_language['language'],
                    language_code=translation_language['language_code'],
                ) for translation_language in self._translation_languages
            )
        )

    def _get_language_description(self, transcript_strings):
        description = '\n'.join(' - {transcript}'.format(transcript=transcript) for transcript in transcript_strings)
        return description if description else 'None'


class Transcript:
    """Your personal transcript handler! 🎭

    >>> transcript = transcript_list.find_transcript(['en'])
    >>> print(transcript.language)
    'English'
    >>> if transcript.is_translatable:
    ...     es_transcript = transcript.translate('es')
    ...     print(es_transcript.language)
    'Spanish'
    """

    def __init__(self, http_client, video_id, url, language, language_code, is_generated, translation_languages):
        """Initialize with all the goodies! 🎁"""
        self._http_client = http_client
        self.video_id = video_id
        self._url = url
        self.language = language
        self.language_code = language_code
        self.is_generated = is_generated
        self.translation_languages = translation_languages
        self._translation_languages_dict = {
            translation_language['language_code']: translation_language['language']
            for translation_language in translation_languages
        }

    def fetch(self, preserve_formatting=False):
        """Get that transcript data! 🎯

        Args:
            preserve_formatting (bool): Keep HTML formatting? Default is nah fam.

        Returns:
            list: That sweet transcript data with text, start time, and duration! 📝
        """
        response = self._http_client.get(self._url, headers={'Accept-Language': 'en-US'})
        return TranscriptParser(preserve_formatting=preserve_formatting).parse(
            _raise_http_errors(response, self.video_id).text,
        )

    def __str__(self):
        """String representation looking clean! 💅"""
        return '{language_code} ("{language}"){translation_description}'.format(
            language=self.language,
            language_code=self.language_code,
            translation_description='[TRANSLATABLE]' if self.is_translatable else ''
        )

    @property
    def is_translatable(self):
        """Can we translate this? 🌍"""
        return len(self.translation_languages) > 0

    def translate(self, language_code):
        """Translate to another language! 🌎

        Args:
            language_code (str): Which language you want fam?

        Returns:
            Transcript: A fresh transcript in your requested language! 🔄

        Raises:
            NotTranslatableError: If we can't translate this one 😢
            TranslationLanguageNotAvailableError: If that language isn't available 🚫
        """
        if not self.is_translatable:
            raise NotTranslatableError(self.video_id)

        if language_code not in self._translation_languages_dict:
            raise TranslationLanguageNotAvailableError(self.video_id)

        return Transcript(
            self._http_client,
            self.video_id,
            '{url}&tlang={language_code}'.format(url=self._url, language_code=language_code),
            self._translation_languages_dict[language_code],
            language_code,
            True,
            [],
        )


class TranscriptParser:
    """Parsing those transcripts like a pro! 🎯

    >>> parser = TranscriptParser(preserve_formatting=True)
    >>> data = parser.parse(xml_data)
    >>> print(data[0])
    {'text': 'Never gonna give you up', 'start': 0.0, 'duration': 4.5}
    """

    _FORMATTING_TAGS = [
        'strong',  # For that extra emphasis 💪
        'em',      # When you need that italic swag 🎨
        'b',       # Bold and beautiful 💯
        'i',       # More italic vibes ✨
        'mark',    # Highlight that text 🌟
        'small',   # Keep it lowkey 🤫
        'del',     # Strike it out ⚡
        'ins',     # Insert new stuff 🆕
        'sub',     # Subscript gang 📉
        'sup',     # Superscript squad 📈
    ]

    def __init__(self, preserve_formatting=False):
        """Get ready to parse with style! 🎨"""
        self._html_regex = self._get_html_regex(preserve_formatting)

    def _get_html_regex(self, preserve_formatting):
        """Get that regex pattern ready! 🎯"""
        if preserve_formatting:
            formats_regex = '|'.join(self._FORMATTING_TAGS)
            formats_regex = r'<\/?(?!\/?(' + formats_regex + r')\b).*?\b>'
            html_regex = re.compile(formats_regex, re.IGNORECASE)
        else:
            html_regex = re.compile(r'<[^>]*>', re.IGNORECASE)
        return html_regex

    def parse(self, plain_data):
        """Parse that XML data into something beautiful! ✨"""
        try:
            return [
                {
                    'text': re.sub(self._html_regex, '', html.unescape(xml_element.text or '')),
                    'start': float(xml_element.attrib['start']),
                    'duration': float(xml_element.attrib.get('dur', '0.0')),
                }
                for xml_element in ElementTree.fromstring(plain_data)
                if xml_element.text is not None
            ]
        except ElementTree.ParseError:
            # If XML parsing fails, try to extract text manually
            return self._fallback_parse(plain_data)

    def _fallback_parse(self, plain_data):
        """Fallback parsing method if XML parsing fails."""
        results = []
        # Try regex pattern matching
        pattern = r'<text start="([^"]+)" dur="([^"]+)"[^>]*>([^<]*)</text>'
        matches = re.findall(pattern, plain_data, re.DOTALL)

        for start, dur, text in matches:
            text = html.unescape(text)
            text = re.sub(self._html_regex, '', text)
            if text.strip():
                results.append({
                    'text': text.strip(),
                    'start': float(start),
                    'duration': float(dur),
                })

        return results


def _raise_http_errors(response, video_id):
    """Handle those HTTP errors with style! 🛠️"""
    try:
        if response.status_code == 429:
            raise TooManyRequestsError(video_id)
        response.raise_for_status()
        return response
    except Exception as error:
        raise YouTubeRequestFailedError(video_id, error)


if __name__ == "__main__":
    # Let's get this party started! 🎉
    from rich import print
    video_url = "https://www.youtube.com/watch?v=dQw4w9WgXcQ"
    transcript = YTTranscriber.get_transcript(video_url, languages=None)
    print("Here's what we got! 🔥")
    print(transcript[:5])



import asyncio
import re
import urllib.parse
from typing import Any, AsyncIterator, Dict, Generator, Literal, Optional, Union, cast

import lxml.html
from curl_cffi.requests import AsyncSession

from webscout import exceptions
from webscout.AIbase import AISearch, SearchResponse
from webscout.litagent import LitAgent
from webscout.scout import Scout

ModeType = Literal["question", "academic", "forums", "wiki", "thinking"]
DetailLevelType = Literal["concise", "detailed", "comprehensive"]


class IAsk(AISearch):
    """A class to interact with the IAsk AI search API.

    IAsk provides a powerful search interface that returns AI-generated responses
    based on web content. It supports both streaming and non-streaming responses,
    as well as different search modes and detail levels.

    Basic Usage:
        >>> from webscout import IAsk
        >>> ai = IAsk()
        >>> # Non-streaming example
        >>> response = ai.search("What is Python?")
        >>> print(response)
        Python is a high-level programming language...

        >>> # Streaming example
        >>> for chunk in ai.search("Tell me about AI", stream=True):
        ...     print(chunk, end="", flush=True)
        Artificial Intelligence is...

        >>> # With specific mode and detail level
        >>> response = ai.search("Climate change", mode="academic", detail_level="detailed")
        >>> print(response)
        Climate change refers to...

    Args:
        timeout (int, optional): Request timeout in seconds. Defaults to 30.
        proxies (dict, optional): Proxy configuration for requests. Defaults to None.
        mode (ModeType, optional): Default search mode. Defaults to "question".
        detail_level (DetailLevelType, optional): Default detail level. Defaults to None.
    """

    def __init__(
        self,
        timeout: int = 30,
        proxies: Optional[dict] = None,
        mode: ModeType = "question",
        detail_level: Optional[DetailLevelType] = None,
    ):
        """Initialize the IAsk API client.

        Args:
            timeout (int, optional): Request timeout in seconds. Defaults to 30.
            proxies (dict, optional): Proxy configuration for requests. Defaults to None.
            mode (ModeType, optional): Default search mode. Defaults to "question".
            detail_level (DetailLevelType, optional): Default detail level. Defaults to None.
        """
        self.timeout = timeout
        self.proxies = proxies or {}
        self.default_mode = mode
        self.default_detail_level = detail_level
        self.api_endpoint = "https://iask.ai"
        self.query_endpoint = "https://iask.ai/q"
        self.last_response = {}
        self.agent = LitAgent()

    def create_url(
        self,
        query: str,
        mode: ModeType = "question",
        detail_level: Optional[DetailLevelType] = None,
    ) -> str:
        """Create a properly formatted URL with mode and detail level parameters.

        Args:
            query (str): The search query.
            mode (ModeType, optional): Search mode. Defaults to "question".
            detail_level (DetailLevelType, optional): Detail level. Defaults to None.

        Returns:
            str: Formatted URL with query parameters.

        Example:
            >>> ai = IAsk()
            >>> url = ai.create_url("Climate change", mode="academic", detail_level="detailed")
            >>> print(url)
            https://iask.ai/q?mode=academic&q=Climate+change&options%5Bdetail_level%5D=detailed
        """
        # Create a dictionary of parameters with flattened structure
        params = {"mode": mode, "q": query}

        # Add detail_level if provided using the flattened format
        if detail_level:
            params["options[detail_level]"] = detail_level

        # Encode the parameters and build the URL
        query_string = urllib.parse.urlencode(params)
        url = f"{self.query_endpoint}?{query_string}"

        return url

    def format_html(self, html_content: str) -> str:
        """Format HTML content into a more readable text format.

        Args:
            html_content (str): The HTML content to format.

        Returns:
            str: Formatted text.
        """
        scout = Scout(html_content, features="html.parser")
        output_lines = []

        for child in scout.find_all(["h1", "h2", "h3", "p", "ol", "ul", "div"]):
            if child.name in ["h1", "h2", "h3"]:
                output_lines.append(f"\n**{child.get_text().strip()}**\n")
            elif child.name == "p":
                text = child.get_text().strip()
                text = re.sub(
                    r"^According to Ask AI & Question AI www\.iAsk\.ai:\s*", "", text
                ).strip()
                # Remove footnote markers
                text = re.sub(r"\[\d+\]\(#fn:\d+ \'see footnote\'\)", "", text)
                output_lines.append(text + "\n")
            elif child.name in ["ol", "ul"]:
                for li in child.find_all("li"):
                    output_lines.append("- " + li.get_text().strip() + "\n")
            elif child.name == "div" and "footnotes" in child.get("class", []):
                output_lines.append("\n**Authoritative Sources**\n")
                for li in child.find_all("li"):
                    link = li.find("a")
                    if link:
                        output_lines.append(f"- {link.get_text().strip()} ({link.get('href')})\n")

        return "".join(output_lines)

    def search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        mode: Optional[ModeType] = None,
        detail_level: Optional[DetailLevelType] = None,
        **kwargs: Any,
    ) -> Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]]:
        """Search using the IAsk API and get AI-generated responses.

        This method sends a search query to IAsk and returns the AI-generated response.
        It supports both streaming and non-streaming modes, as well as raw response format.

        Args:
            prompt (str): The search query or prompt to send to the API.
            stream (bool, optional): If True, yields response chunks as they arrive.
                                   If False, returns complete response. Defaults to False.
            raw (bool, optional): If True, returns raw response dictionaries with 'text' key.
                                If False, returns Response objects that convert to text automatically.
                                Defaults to False.
            mode (ModeType, optional): Search mode to use. Defaults to None (uses instance default).
            detail_level (DetailLevelType, optional): Detail level to use. Defaults to None (uses instance default).

        Returns:
            Union[Response, Generator[Union[Dict[str, str], Response], None, None]]:
                - If stream=False: Returns complete response as Response object
                - If stream=True: Yields response chunks as either Dict or Response objects

        Raises:
            APIConnectionError: If the API request fails

        Examples:
            Basic search:
            >>> ai = IAsk()
            >>> response = ai.search("What is Python?")
            >>> print(response)
            Python is a programming language...

            Streaming response:
            >>> for chunk in ai.search("Tell me about AI", stream=True):
            ...     print(chunk, end="")
            Artificial Intelligence...

            Raw response format:
            >>> for chunk in ai.search("Hello", stream=True, raw=True):
            ...     print(chunk)
            {'text': 'Hello'}
            {'text': ' there!'}

            With specific mode and detail level:
            >>> response = ai.search("Climate change", mode="academic", detail_level="detailed")
            >>> print(response)
            Climate change refers to...
        """
        search_mode = cast(ModeType, mode or self.default_mode)
        search_detail_level = cast(
            Optional[DetailLevelType], detail_level or self.default_detail_level
        )

        # For non-streaming, run the async search and return the complete response
        if not stream:
            # Create a new event loop for this request
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                result = loop.run_until_complete(
                    self._async_search(prompt, False, raw, search_mode, search_detail_level)
                )
                return cast(
                    Union[
                        SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]
                    ],
                    result,
                )
            finally:
                loop.close()
        buffer = ""

        def sync_generator():
            nonlocal buffer

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            try:
                # Get the async generator
                async_gen_coro = self._async_search(
                    prompt, True, raw, search_mode, search_detail_level
                )
                async_gen = loop.run_until_complete(async_gen_coro)

                # Process chunks one by one
                if hasattr(async_gen, "__anext__"):
                    async_iterator = cast(AsyncIterator, async_gen)
                    while True:
                        try:
                            # Get the next chunk
                            chunk_coro = async_iterator.__anext__()
                            chunk = loop.run_until_complete(chunk_coro)

                            # Update buffer and yield the chunk
                            if isinstance(chunk, dict) and "text" in chunk:
                                buffer += chunk["text"]
                            elif isinstance(chunk, SearchResponse):
                                buffer += chunk.text
                            else:
                                buffer += str(chunk)

                            yield chunk
                        except StopAsyncIteration:
                            break
                        except Exception as e:
                            print(f"Error in generator: {e}")
                            break
                elif isinstance(async_gen, SearchResponse):
                    yield async_gen
                else:
                    yield str(async_gen)
            finally:
                # Store the final response and close the loop
                self.last_response = {"text": buffer}
                loop.close()

        return cast(
            Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]],
            sync_generator(),
        )

    def _extract_answer(self, html_content: str) -> str:
        """Extract and format the answer HTML from the results page."""
        etree = lxml.html.fromstring(html_content)
        text_nodes = etree.xpath('//*[@id="text"]')
        if not text_nodes:
            raise exceptions.APIConnectionError("No answer content found in iAsk response.")
        text_node = text_nodes[0]
        text_html = lxml.html.tostring(text_node, encoding="unicode", method="html")
        return self.format_html(text_html).strip()

    def _iter_chunks(self, text: str) -> Generator[str, None, None]:
        """Yield a response string in readable chunks for streaming."""
        for line in text.splitlines(keepends=True):
            if line.strip() == "":
                yield line
                continue
            for chunk in re.findall(r".{1,800}(?:\s+|$)", line):
                yield chunk

    async def _async_search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        mode: ModeType = "question",
        detail_level: Optional[DetailLevelType] = None,
    ) -> Union[SearchResponse, str, AsyncIterator[Union[str, Dict[str, str], SearchResponse]]]:
        """Internal async implementation of the search method."""

        async def fetch_answer() -> str:
            timeout = self.timeout
            params = {"mode": mode, "q": prompt}
            if detail_level:
                params["options[detail_level]"] = detail_level

            headers = {
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "User-Agent": self.agent.random(),
            }

            async with AsyncSession() as session:
                response = await session.get(
                    self.query_endpoint,
                    params=params,
                    headers=headers,
                    proxies=self.proxies or None,
                    timeout=timeout,
                )
            if response.status_code != 200:
                raise exceptions.APIConnectionError(
                    "Failed to generate response - "
                    f"({response.status_code}, {response.reason}) - {response.text}"
                )
            return self._extract_answer(response.text)

        # For non-streaming, collect all chunks and return a single response
        if not stream:
            buffer = await fetch_answer()
            self.last_response = SearchResponse(buffer)
            return buffer if raw else self.last_response

        # For streaming, create an async generator that yields chunks
        async def process_stream():
            buffer = ""
            text = await fetch_answer()
            for chunk in self._iter_chunks(text):
                buffer += chunk
                if raw:
                    yield chunk
                else:
                    yield SearchResponse(chunk)
            self.last_response = SearchResponse(buffer)

        # Return the async generator
        return process_stream()


if __name__ == "__main__":
    ai = IAsk()
    response = ai.search("What is Python?", stream=True, raw=True)
    if hasattr(response, "__iter__") and not isinstance(response, (str, SearchResponse)):
        for chunk in response:
            print(chunk, end="", flush=True)
    else:
        print(response)



"""Yandex search engine."""

from __future__ import annotations

from collections.abc import Mapping
from random import SystemRandom
from typing import Any, Optional

from ..base import BaseSearchEngine
from ..results import TextResult

random = SystemRandom()


class Yandex(BaseSearchEngine[TextResult]):
    """Yandex search engine."""

    name = "yandex"
    category = "text"
    provider = "yandex"

    search_url = "https://yandex.com/search/"
    search_method = "GET"

    items_xpath = "//li[contains(@class, 'serp-item')]"
    elements_xpath: Mapping[str, str] = {
        "title": ".//h2//text()",
        "href": ".//h2/a/@href",
        "body": ".//div[contains(@class, 'text-container')]//text()",
    }

    def build_payload(
        self, query: str, region: str, safesearch: str, timelimit: str | None, page: int = 1, **kwargs: Any
    ) -> dict[str, Any]:
        """Build a payload for the search request."""
        safesearch_base = {"on": "1", "moderate": "0", "off": "0"}
        payload = {
            "text": query,
            "family": safesearch_base[safesearch.lower()],
        }
        if page > 1:
            payload["p"] = str(page - 1)
        return payload

    def run(self, *args, **kwargs) -> list[TextResult]:
        """Run text search on Yandex.

        Args:
            keywords: Search query.
            region: Region code.
            safesearch: Safe search level.
            max_results: Maximum number of results (ignored for now).

        Returns:
            List of TextResult objects.
        """
        keywords = args[0] if args else kwargs.get("keywords")
        if keywords is None:
            keywords = ""
        region = args[1] if len(args) > 1 else kwargs.get("region", "us-en")
        safesearch = args[2] if len(args) > 2 else kwargs.get("safesearch", "moderate")
        max_results = args[3] if len(args) > 3 else kwargs.get("max_results")

        results = self.search(query=keywords, region=region, safesearch=safesearch)
        if results and max_results:
            results = results[:max_results]
        return results or []



from __future__ import annotations

from typing import List, Optional
from urllib.parse import urlencode

from webscout.search.results import TextResult

from .base import YepBase


class YepSearch(YepBase):
    name = "yep"
    category = "text"
    def run(self, *args, **kwargs) -> List[TextResult]:
        keywords = args[0] if args else kwargs.get("keywords")
        region = args[1] if len(args) > 1 else kwargs.get("region", "all")
        safesearch = args[2] if len(args) > 2 else kwargs.get("safesearch", "moderate")
        max_results = args[3] if len(args) > 3 else kwargs.get("max_results")

        safe_search_map = {
            "on": "on",
            "moderate": "moderate",
            "off": "off"
        }
        safe_setting = safe_search_map.get(safesearch.lower(), "moderate")

        params = {
            "client": "web",
            "gl": region,
            "limit": str(max_results) if max_results else "10",
            "no_correct": "false",
            "q": keywords,
            "safeSearch": safe_setting,
            "type": "web"
        }

        url = f"{self.base_url}?{urlencode(params)}"
        try:
            response = self.session.get(url)
            response.raise_for_status()
            raw_results = response.json()

            formatted_results = self.format_results(raw_results)

            if max_results:
                return formatted_results[:max_results]
            return formatted_results
        except Exception as e:
            resp = getattr(e, 'response', None)
            if resp is not None:
                 raise Exception(f"Yep search failed with status {resp.status_code}: {str(e)}")
            else:
                 raise Exception(f"Yep search failed: {str(e)}")

    def format_results(self, raw_results: dict) -> List[TextResult]:
        formatted_results = []

        if not raw_results or len(raw_results) < 2:
            return formatted_results

        results = raw_results[1].get('results', [])

        for result in results:
            formatted_result = TextResult(
                title=self._remove_html_tags(result.get("title", "")),
                href=result.get("url", ""),
                body=self._remove_html_tags(result.get("snippet", ""))
            )

            formatted_results.append(formatted_result)

        return formatted_results

    def _remove_html_tags(self, text: str) -> str:
        result = ""
        in_tag = False

        for char in text:
            if char == '<':
                in_tag = True
            elif char == '>':
                in_tag = False
            elif not in_tag:
                result += char

        replacements = {
            '&nbsp;': ' ',
            '&amp;': '&',
            '&lt;': '<',
            '&gt;': '>',
            '&quot;': '"',
            '&apos;': "'",
        }

        for entity, replacement in replacements.items():
            result = result.replace(entity, replacement)

        return result.strip()




"""Prompt optimization utilities."""

import os
import platform
import subprocess
from typing import Optional


class Optimizers:
    """
    >>> Optimizers.coder("write a hello world")
    Returns optimized prompt for code generation
    """

    @staticmethod
    def coder(prompt: str) -> str:
        """Unified optimizer for both code and shell commands."""
        # Get system info for shell commands
        operating_system: str = ""
        if platform.system() == "Windows":
            operating_system = "Windows"
        elif platform.system() == "Darwin":
            operating_system = "MacOS"
        elif platform.system() == "Linux":
            try:
                result: str = subprocess.check_output(["lsb_release", "-si"]).decode().strip()
                operating_system = f"Linux/{result}" if result else "Linux"
            except Exception:
                operating_system = "Linux"
        else:
            operating_system = platform.system()

        # Get shell info
        shell_name: str = "/bin/sh"
        if platform.system() == "Windows":
            shell_name = "powershell.exe" if os.getenv("PSModulePath") else "cmd.exe"
        else:
            shell_env: Optional[str] = os.getenv("SHELL")
            if shell_env:
                shell_name = shell_env

        return (
            f"""<system_context>
<role>
  Your Role: You are a code generation expert. Analyze the request and provide appropriate output.
  If the request starts with '!' or involves system/shell operations, provide a shell command.
  Otherwise, provide Python code.
</role>
<rules>
   RULES:
     - Provide ONLY code/command output without any description or markdown
     - For shell commands:
         - Target OS: {operating_system}
         - Shell: {shell_name}
         - Combine multiple steps when possible
         - Use appropriate flags for safety and clarity
     - For Python code:
        - Include necessary imports
        - Handle errors appropriately
        - Follow PEP 8 style
        - Use type hints where appropriate
        - Include docstrings for functions
     - If details are missing, use most logical implementation
     - No warnings, descriptions, or explanations
     - For complex tasks, break down into smaller logical steps
</rules>
<request>
     Request: {prompt}
</request>
<output>
    Output:
</output>
</system_context>"""
        )



import json
import time
from typing import Any
from urllib.error import HTTPError
from urllib.request import Request, urlopen

try:
    from webscout.litagent.agent import LitAgent
    _USER_AGENT_GENERATOR = LitAgent()
except ImportError:
    _USER_AGENT_GENERATOR = None

class GitError(Exception):
    """Base exception for GitHub API errors"""
    pass

class RateLimitError(GitError):
    """Raised when hitting GitHub API rate limits"""
    pass

class NotFoundError(GitError):
    """Raised when resource is not found"""
    pass

class RequestError(GitError):
    """Raised for general request errors"""
    pass


def request(url: str, retry_attempts: int = 3) -> Any:
    """
    Send a request to GitHub API with retry mechanism

    Args:
        url: GitHub API endpoint URL
        retry_attempts: Number of retry attempts

    Returns:
        Parsed JSON response

    Raises:
        NotFoundError: If resource not found
        RateLimitError: If rate limited
        RequestError: For other request errors
    """
    headers = {
        "User-Agent": _USER_AGENT_GENERATOR.random() if _USER_AGENT_GENERATOR else "webscout-gitapi/1.0",
        "Accept": "application/vnd.github+json"
    }

    for attempt in range(retry_attempts):
        try:
            req = Request(url, headers=headers)
            response = urlopen(req, timeout=30)
            data = response.read().decode('utf-8')
            try:
                return json.loads(data)
            except json.JSONDecodeError as json_err:
                raise RequestError(f"Invalid JSON response from {url}: {str(json_err)}")

        except HTTPError as e:
            if e.code == 404:
                raise NotFoundError(f"Resource not found: {url}")
            if e.code == 429:
                if attempt < retry_attempts - 1:
                    # Wait before retrying on rate limit
                    time.sleep(2 ** attempt)  # Exponential backoff
                    continue
                raise RateLimitError(f"Rate limited after {retry_attempts} attempts")
            if e.code == 403:
                raise RequestError("Forbidden: Check your authentication token")
            if attempt == retry_attempts - 1:
                raise RequestError(f"HTTP Error {e.code}: {e.reason}")
            # Wait before retrying on other HTTP errors
            time.sleep(1)

        except Exception as e:
            if attempt == retry_attempts - 1:
                raise RequestError(f"Request failed: {str(e)}")
            time.sleep(1)

    raise RequestError(f"Request to {url} failed after {retry_attempts} attempts")



import json
import random
from typing import Any, Dict, Generator, List, Optional, Union
from uuid import uuid4

from curl_cffi.requests import Session

from webscout import LitAgent, exceptions
from webscout.AIbase import AISearch, SearchResponse


class Liner(AISearch):
    """A class to interact with the Liner AI search API.

    Liner provides a powerful search interface that returns AI-generated responses
    based on web content. It supports both streaming and non-streaming responses.

    Basic Usage:
        >>> from webscout import Liner
        >>> ai = Liner(cookies_path="cookies.json")
        >>> # Non-streaming example
        >>> response = ai.search("What is Python?")
        >>> print(response)
        Python is a high-level programming language...

        >>> # Streaming example
        >>> for chunk in ai.search("Tell me about AI", stream=True):
        ...     print(chunk, end="", flush=True)
        Artificial Intelligence is...

        >>> # Raw response format
        >>> for chunk in ai.search("Hello", stream=True, raw=True):
        ...     print(chunk)
        {'text': 'Hello'}
        {'text': ' there!'}

    Args:
        cookies_path (str): Path to the cookies JSON file
        timeout (int, optional): Request timeout in seconds. Defaults to 30.
        proxies (dict, optional): Proxy configuration for requests. Defaults to None.
        deep_search (bool, optional): Enable deep research mode. Defaults to True.
        reasoning_mode (bool, optional): Enable reasoning mode. Defaults to False.
    """

    def __init__(
        self,
        cookies_path: str,
        timeout: int = 600,
        proxies: Optional[dict] = None,
        deep_search: bool = True,
        reasoning_mode: bool = False,
    ):
        """Initialize the Liner API client.

        Args:
            cookies_path (str): Path to the cookies JSON file
            timeout (int, optional): Request timeout in seconds. Defaults to 30.
            proxies (dict, optional): Proxy configuration for requests. Defaults to None.
            deep_search (bool, optional): Enable deep research mode. Defaults to True.
            reasoning_mode (bool, optional): Enable reasoning mode. Defaults to False.
        """
        self.session = Session()
        self.chat_endpoint = "https://getliner.com/lisa/v1/answer"
        self.stream_chunk_size = 64
        self.timeout = timeout
        self.last_response = {}
        self.cookies_path = cookies_path
        self.deep_search = deep_search
        self.reasoning_mode = reasoning_mode

        # Generate random IDs
        self.space_id = random.randint(10000000, 99999999)
        self.thread_id = random.randint(10000000, 99999999)
        self.user_message_id = random.randint(100000000, 999999999)
        self.user_id = random.randint(1000000, 9999999)

        self.headers = {
            "accept": "text/event-stream",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://getliner.com",
            "referer": f"https://getliner.com/search/s/{self.space_id}/t/t_{uuid4()}?mode=temp&msg-entry-type=main&build-id=kwJaNRjnCKjh7PijZgqV2",
            "sec-ch-ua": '"Chromium";v="134", "Not:A-Brand";v="24", "Microsoft Edge";v="134"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-gpc": "1",
            "user-agent": LitAgent().random()
        }

        # Load cookies from JSON file
        self.cookies = self._load_cookies()
        if not self.cookies:
            raise ValueError("Failed to load cookies from file")

        # Set headers and cookies in session
        self.session.headers.update(self.headers)
        self.session.cookies.update(self.cookies)
        if proxies:
            self.session.proxies.update(proxies) or {}

    def _load_cookies(self) -> Optional[Dict[str, str]]:
        """Load cookies from a JSON file.

        Returns:
            Optional[Dict[str, str]]: Dictionary of cookies if successful, None otherwise
        """
        try:
            with open(self.cookies_path, 'r') as f:
                cookies_data = json.load(f)
            return {cookie['name']: cookie['value'] for cookie in cookies_data}
        except FileNotFoundError:
            print(f"Error: {self.cookies_path} file not found!")
            return None
        except json.JSONDecodeError:
            print(f"Error: Invalid JSON format in {self.cookies_path}!")
            return None
        except KeyError:
            print(f"Error: Invalid cookie format in {self.cookies_path}! Each cookie must have 'name' and 'value' keys.")
            return None

    def search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None], List[Any], Dict[str, Any], str]:
        """Search using the Liner API and get AI-generated responses.

        Args:
            prompt (str): The search query or prompt to send to the API.
            stream (bool, optional): If True, yields response chunks as they arrive.
                                   If False, returns complete response. Defaults to False.
            raw (bool, optional): If True, returns raw response dictionaries with 'text' key.
                                If False, returns SearchResponse objects that convert to text automatically.
                                Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None], List[Any], Dict[str, Any], str]:
                - If stream=False: Returns complete response
                - If stream=True: Yields response chunks as they arrive

        Raises:
            APIConnectionError: If the API request fails
        """
        payload = {
            "spaceId": self.space_id,
            "threadId": self.thread_id,
            "userMessageId": self.user_message_id,
            "userId": self.user_id,
            "query": prompt,
            "agentId": "liner",
            "platform": "web",
            "regenerate": False,
            "showReferenceChunks": True,
            "mode": "general",
            "answerMode": "search",
            "isReasoningMode": self.reasoning_mode,
            "experimentId": random.randint(80, 90),
            "modelType": "liner",
            "experimentVariants": [],
            "isDeepResearchMode": self.deep_search
        }

        def for_stream():
            try:
                with self.session.post(
                    self.chat_endpoint,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                ) as response:
                    if not response.ok:
                        raise exceptions.APIConnectionError(
                            f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                        )

                    current_reasoning = ""
                    current_answer = ""

                    for line in response.iter_lines(decode_unicode=True):
                        if line == "event:finish_answer":
                            break

                        if line.startswith('data:'):
                            try:
                                data = json.loads(line[5:])  # Remove 'data:' prefix

                                # Handle reasoning updates if enabled
                                if self.reasoning_mode and 'reasoning' in data:
                                    current_reasoning += data['reasoning']
                                    if raw:
                                        yield {"text": data['reasoning']}
                                    else:
                                        yield SearchResponse(data['reasoning'])

                                # Handle answer updates
                                if 'answer' in data:
                                    current_answer += data['answer']
                                    if raw:
                                        yield {"text": data['answer']}
                                    else:
                                        yield SearchResponse(data['answer'])

                            except json.JSONDecodeError:
                                continue

            except Exception as e:
                raise exceptions.APIConnectionError(f"Request failed: {e}")

        def for_non_stream():
            full_response = ""
            for chunk in for_stream():
                if raw:
                    full_response = chunk
                    break
                else:
                    full_response += str(chunk)

            if raw:
                return full_response
            else:
                self.last_response = SearchResponse(full_response)
                return self.last_response

        return for_stream() if stream else for_non_stream()


if __name__ == "__main__":
    from typing import Generator

    from rich import print

    ai = Liner(cookies_path="cookies.json")
    response = ai.search(input(">>> "), stream=True, raw=False)
    if isinstance(response, Generator):
        for chunk in response:
            print(chunk, end="", flush=True)
    else:
        print(response)



import importlib.metadata
import os
import sys
import tempfile
import time
from pathlib import Path
from typing import Any, Dict, Optional

from curl_cffi.requests import Session
from packaging import version

try:
    from rich.console import Console
    from rich.panel import Panel
    from rich.text import Text

    HAS_RICH = True
except ImportError:

    class _Dummy:
        def __init__(self, *_args: Any, **_kwargs: Any) -> None:
            pass

        def print(self, *_args: Any, **_kwargs: Any) -> None:
            pass

        @staticmethod
        def assemble(*_args: Any, **_kwargs: Any) -> str:
            return ""

    Console = _Dummy  # type: ignore
    Panel = _Dummy  # type: ignore
    Text = _Dummy  # type: ignore
    HAS_RICH = False

PYPI_URL = "https://pypi.org/pypi/webscout/json"
YOUTUBE_URL = "https://youtube.com/@OEvortex"
GITHUB_URL = "https://github.com/OEvortex/Webscout"
CACHE_FILE = Path(tempfile.gettempdir()) / "webscout_update_check.cache"

session = Session()


def get_installed_version() -> str:
    """Get the currently installed version of webscout."""
    try:
        from .version import __version__

        return __version__
    except (ImportError, ValueError):
        pass

    try:
        import webscout.version

        return webscout.version.__version__
    except (ImportError, AttributeError):
        pass

    try:
        return importlib.metadata.version("webscout")
    except importlib.metadata.PackageNotFoundError:
        return "0.0.0"


def get_pypi_versions() -> Dict[str, Optional[str]]:
    """Get stable and latest versions from PyPI."""
    try:
        response = session.get(PYPI_URL, timeout=3)
        response.raise_for_status()
        data = response.json()

        stable = data.get("info", {}).get("version")

        releases = data.get("releases", {}).keys()
        parsed_versions = []
        for v in releases:
            try:
                parsed_versions.append(version.parse(v))
            except Exception:
                continue

        all_versions = sorted(parsed_versions)
        latest = str(all_versions[-1]) if all_versions else stable

        return {"stable": stable, "latest": latest}
    except Exception:
        return {"stable": None, "latest": None}


def should_check(force: bool = False) -> bool:
    """Check if we should perform an update check based on cache."""
    if os.environ.get("WEBSCOUT_NO_UPDATE"):
        return False

    if force:
        return True

    try:
        if not CACHE_FILE.exists():
            return True

        last_check = float(CACHE_FILE.read_text().strip())
        if time.time() - last_check > 43200:
            return True
    except Exception:
        return True
    return False


def mark_checked() -> Any:
    """Mark the current time as the last update check."""
    try:
        CACHE_FILE.write_text(str(time.time()))
    except Exception:
        pass


def is_venv() -> bool:
    """Check if we are running inside a virtual environment."""
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


def get_env_name() -> str:
    """Get the name of the current environment."""
    if is_venv():
        return os.path.basename(sys.prefix)
    return "Global"


def format_update_message(current: str, new: str, utype: str) -> str:
    """Format the update message with colors and style."""
    cmd = "pip install -U webscout"
    if utype == "Pre-release" or version.parse(new).is_prerelease:
        cmd = "pip install -U --pre webscout"

    if HAS_RICH:
        from io import StringIO

        capture = StringIO()
        console = Console(file=capture, force_terminal=True, width=80)

        content = Text.assemble(
            ("A new ", "white"),
            (f"{utype} ", "bold yellow" if utype == "Pre-release" else "bold green"),
            ("version of Webscout is available!\n\n", "white"),
            ("Current:     ", "white"),
            (f"{current}", "bold red"),
            ("\n", ""),
            ("Latest:      ", "white"),
            (f"{new}", "bold green"),
            ("\n\n", ""),
            ("To update, run: ", "white"),
            (f"{cmd}", "bold cyan"),
            ("\n\n", ""),
            ("Subscribe to my YouTube: ", "dim"),
            (f"{YOUTUBE_URL}", "dim cyan"),
            ("\n", ""),
            ("Star on GitHub: ", "dim"),
            (f"{GITHUB_URL}", "dim cyan"),
        )

        panel = Panel(
            content,
            title="[bold magenta]Update Available[/bold magenta]",
            border_style="bright_blue",
            expand=False,
            padding=(1, 2),
        )
        console.print(panel)
        return capture.getvalue()
    else:
        return (
            f"\n\033[1;36m[ Webscout Update ]\033[0m\n"
            f"New {utype} version available: \033[1;32m{new}\033[0m (Current: \033[1;31m{current}\033[0m)\n"
            f"Run: \033[1;33m{cmd}\033[0m to update.\n"
            f"\033[1;32mYouTube: {YOUTUBE_URL}\033[0m\n"
        )


def format_dev_message(current: str, latest: str) -> str:
    """Format the message for development versions."""
    if HAS_RICH:
        from io import StringIO

        capture = StringIO()
        console = Console(file=capture, force_terminal=True, width=80)

        content = Text.assemble(
            ("You are running a ", "white"),
            ("Development Version", "bold yellow"),
            ("\n\nLocal Version: ", "white"),
            (f"{current}", "bold cyan"),
            ("\n", ""),
            ("Latest PyPI:   ", "white"),
            (f"{latest}", "bold green"),
            ("\n\n", ""),
            ("YouTube: ", "dim"),
            (f"{YOUTUBE_URL}", "dim cyan"),
        )

        panel = Panel(
            content,
            title="[bold blue]Webscout Dev Mode[/bold blue]",
            border_style="yellow",
            expand=False,
            padding=(1, 2),
        )
        console.print(panel)
        return capture.getvalue()
    else:
        return (
            f"\n\033[1;33m[ Webscout Info ]\033[0m\n"
            f"You're running a development version (\033[1;36m{current}\033[0m)\n"
            f"Latest stable release: \033[1;32m{latest}\033[0m\n"
            f"\033[1;32mYouTube: {YOUTUBE_URL}\033[0m\n"
        )


def check_for_updates(force: bool = False) -> Optional[str]:
    """
    Check if a newer version of Webscout is available.

    Args:
        force (bool): If True, ignore cache and force check.

    Returns:
        Optional[str]: Formatted update message or None.
    """
    if not sys.stdout.isatty() and not force:
        return None

    if not should_check(force):
        return None

    try:
        installed_str = get_installed_version()
        installed_v = version.parse(installed_str)

        pypi = get_pypi_versions()
        mark_checked()

        stable_ver = pypi.get("stable")
        latest_ver = pypi.get("latest")

        if not stable_ver or not latest_ver:
            return None

        latest_stable_str = stable_ver
        latest_stable_v = version.parse(latest_stable_str)

        latest_any_str = latest_ver
        latest_any_v = version.parse(latest_any_str)

        is_prerelease = installed_v.is_prerelease

        if is_prerelease:
            if installed_v < latest_any_v:
                utype = "Pre-release" if latest_any_v.is_prerelease else "Stable"
                return format_update_message(installed_str, latest_any_str, utype)
            elif installed_v > latest_any_v:
                return format_dev_message(installed_str, latest_any_str)
        else:
            if installed_v < latest_stable_v:
                return format_update_message(installed_str, latest_stable_str, "Stable")
            elif installed_v > latest_stable_v:
                return format_dev_message(installed_str, latest_stable_str)

    except Exception:
        pass

    return None


if __name__ == "__main__":
    msg = check_for_updates(force=True)
    if msg:
        print(msg)



import json
from typing import Any, Dict, Generator, List, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream
from webscout.litagent import LitAgent as Lit
from webscout.model_fetcher import BackgroundModelFetcher


class Sambanova(Provider):
    """
    A class to interact with the Sambanova API.
    """
    required_auth = True
    AVAILABLE_MODELS = [
        "DeepSeek-R1",
        "DeepSeek-V3",
        "Meta-Llama-3.3-70B-Instruct",
        "Meta-Llama-3.1-8B-Instruct",
        "Meta-Llama-3.1-70B-Instruct",
        "Meta-Llama-3.1-405B-Instruct",
        "Qwen2.5-72B-Instruct",
        "Qwen2.5-Coder-32B-Instruct"
    ]
    # Background model fetcher
    _model_fetcher = BackgroundModelFetcher()

    @classmethod
    def get_models(cls, api_key: Optional[str] = None):
        """Fetch available models from Sambanova API."""
        if not api_key:
            return cls.AVAILABLE_MODELS

        try:
            temp_session = Session()
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {api_key}",
            }

            response = temp_session.get(
                "https://api.sambanova.ai/v1/models",
                headers=headers,
                impersonate="chrome120"
            )

            if response.status_code == 200:
                data = response.json()
                if "data" in data and isinstance(data["data"], list):
                    new_models = [model['id'] for model in data['data'] if 'id' in model]
                    if new_models:
                        return new_models

        except Exception:
            pass
        return cls.AVAILABLE_MODELS

    def __init__(
        self,
        api_key: Optional[str] = None,
        is_conversation: bool = True,
        max_tokens: int = 4096,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "Meta-Llama-3.1-8B-Instruct",
        system_prompt: str = "You are a helpful AI assistant.",
    ):
        """
        Initializes the Sambanova API with given parameters.
        """
        self.api_key = api_key
        self.model = model
        self.system_prompt = system_prompt
        self.timeout = timeout

        # Start background model fetch (non-blocking)
        self._model_fetcher.fetch_async(
            provider_name="Sambanova",
            fetch_func=lambda: self.get_models(api_key),
            fallback_models=self.AVAILABLE_MODELS,
            timeout=10,
        )

        # Initialize curl_cffi Session
        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.last_response = {}

        self.headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "User-Agent": Lit().random(),
        }

        # Update curl_cffi session headers and proxies
        self.session = Session()
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)
        self.system_prompt = system_prompt

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        if act:
            self.conversation.intro = AwesomePrompts().get_act(cast(Union[str, int], act), default=self.conversation.intro, case_insensitive=True
            ) or self.conversation.intro
        elif intro:
            self.conversation.intro = intro

        # Configure the API base URL
        self.base_url = "https://api.sambanova.ai/v1/chat/completions"

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Response:
        """Chat with AI using the Sambanova API."""
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(
                    f"Optimizer is not one of {list(self.__available_optimizers)}"
                )

        payload = {
            "model": self.model,
            "messages": [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": conversation_prompt},
            ],
            "max_tokens": self.max_tokens_to_sample,
            "stream": stream
        }

        if tools:
            payload["tools"] = tools
        if tool_choice:
            payload["tool_choice"] = tool_choice

        def for_stream():
            try:
                # Use curl_cffi session post with impersonate
                response = self.session.post(
                    self.base_url,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome120"
                )
                response.raise_for_status()

                streaming_text = ""
                processed_stream = sanitize_stream(
                    data=response.iter_lines(),
                    intro_value="data:",
                    to_json=True,
                    skip_markers=["[DONE]"],
                    content_extractor=lambda chunk: chunk.get('choices', [{}])[0].get('delta') if isinstance(chunk, dict) else None,
                    yield_raw_on_error=False,
                    raw=raw
                )

                for delta in processed_stream:
                    if isinstance(delta, dict):
                        if 'content' in delta and delta['content'] is not None:
                            content = delta['content']
                            if raw:
                                yield content
                            else:
                                streaming_text += content
                                yield {"text": content}
                        elif 'tool_calls' in delta:
                            tool_calls = delta['tool_calls']
                            if raw:
                                yield json.dumps(tool_calls)
                            else:
                                yield {"tool_calls": tool_calls}

                self.last_response.update({"text": streaming_text})
                if streaming_text:
                    self.conversation.update_chat_history(prompt, streaming_text)

            except CurlError as e:
                raise exceptions.ProviderConnectionError(f"Request failed (CurlError): {e}") from e
            except Exception as e:
                raise exceptions.ProviderConnectionError(f"Request failed ({type(e).__name__}): {e}") from e

        def for_non_stream():
            try:
                payload["stream"] = False
                response = self.session.post(
                    self.base_url,
                    json=payload,
                    timeout=self.timeout,
                    impersonate="chrome120"
                )
                response.raise_for_status()
                resp_json = response.json()

                if 'choices' in resp_json and len(resp_json['choices']) > 0:
                    choice = resp_json['choices'][0]
                    message = choice.get('message', {})
                    content = message.get('content', '')
                    tool_calls = message.get('tool_calls')

                    result = {}
                    if content:
                        result["text"] = content
                    if tool_calls:
                        result["tool_calls"] = tool_calls

                    self.last_response = result
                    self.conversation.update_chat_history(prompt, content or "")

                    if raw:
                        return content if content else (json.dumps(tool_calls) if tool_calls else "")
                    return result
                else:
                    return {}

            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Non-stream request failed: {e}") from e

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """Generate response `str`"""
        raw = kwargs.get("raw", False)
        tools = kwargs.get("tools")
        tool_choice = kwargs.get("tool_choice")

        def for_stream_chat():
             # ask() yields dicts or strings when streaming
             gen = self.ask(
                 prompt, stream=True, raw=raw,
                 optimizer=optimizer, conversationally=conversationally,
                 tools=tools, tool_choice=tool_choice
             )
             for response_dict in gen:
                 if raw:
                     yield cast(str, response_dict)
                 else:
                     yield self.get_message(cast(Response, response_dict))

        def for_non_stream_chat():
             # ask() returns dict or str when not streaming
             response_data = self.ask(
                 prompt,
                 stream=False,
                 raw=raw,
                 optimizer=optimizer,
                 conversationally=conversationally,
                 tools=tools,
                 tool_choice=tool_choice
             )
             if raw:
                 return cast(str, response_data)
             return self.get_message(response_data)

        return for_stream_chat() if stream else for_non_stream_chat()

    def get_message(self, response: Response) -> str:
        """
        Retrieves a clean message from the provided response.

        Args:
            response: The raw response data.

        Returns:
            str: The extracted message.
        """
        if isinstance(response, str):
            return response
        elif isinstance(response, dict):
            resp_dict = cast(Dict[str, Any], response)
            if "text" in resp_dict:
                return cast(str, resp_dict["text"])
            elif "tool_calls" in resp_dict:
                return json.dumps(resp_dict["tool_calls"])
        return str(response)

if __name__ == "__main__":
    # Ensure curl_cffi is installed
    from rich import print
    ai = Sambanova(api_key='')
    response = ai.chat(input(">>> "), stream=True)
    if hasattr(response, "__iter__") and not isinstance(response, (str, bytes)):
        for chunk in response:
            print(chunk, end="", flush=True)
    else:
        print(response)



import json
import random
import re
import string
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union

import cloudscraper  # type: ignore
from curl_cffi.requests import Session
from rich import print

from webscout.litagent.agent import LitAgent

# Import base classes and utility structures
from webscout.Provider.Openai_comp.base import (
    BaseChat,
    BaseCompletions,
    OpenAICompatibleProvider,
    SimpleModelList,
)
from webscout.Provider.Openai_comp.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
)


# --- ChutesAI API Key Auto-Generator ---
def generate_chutesai_api_key():
    url = "https://chutes.ai/auth/start?/create"
    def generate_username(length=8):
        return ''.join(random.choices(string.ascii_letters, k=length))
    username = generate_username()
    agent = LitAgent()
    fingerprint = agent.generate_fingerprint("chrome")
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Accept": "application/json",
        "User-Agent": fingerprint["user_agent"],
        "Accept-Encoding": "gzip, deflate, br, zstd",
        "Accept-Language": fingerprint["accept_language"],
        "DNT": "1",
        "Origin": "https://chutes.ai",
        "Referer": "https://chutes.ai/auth/start",
        "Sec-Ch-Ua": fingerprint["sec_ch_ua"],
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": fingerprint["platform"],
        "X-Sveltekit-Action": "true"
    }
    data = {
        "username": username,
        "coldkey": "hotkey",
        "__superform_id": "xpsmbd"
    }
    scraper = cloudscraper.create_scraper()
    response = scraper.post(url, headers=headers, data=data)
    print(f"[bold green]Status:[/] {response.status_code}")

    # Ensure response is decoded as UTF-8
    response.encoding = 'utf-8'

    try:
        resp_json = response.json()
    except Exception:
        try:
            # Try to decode the response text with UTF-8 explicitly
            decoded_text = response.content.decode('utf-8', errors='replace')
            print(decoded_text)
        except Exception:
            print("Failed to decode response content")
        return None
    print(resp_json)
    # Extract the api_key using regex from the 'data' field
    if 'data' in resp_json:
        api_key_match = re.search(r'(cpk_[a-zA-Z0-9.]+)', resp_json['data'])
        if api_key_match:
            api_key = api_key_match.group(1)
            print(f"[bold yellow]Auto-generated ChutesAI API Key:[/] {api_key}")
            return api_key
        else:
            print("[red]API key not found in response data.")
    return None

# --- ChutesAI Client ---

class Completions(BaseCompletions):
    def __init__(self, client: 'ChutesAI'):
        self._client = client

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = 1024,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        **kwargs: Any
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Creates a model response for the given chat conversation.
        Mimics openai.chat.completions.create
        """
        payload = {
            "model": model,
            "messages": messages,
            "max_tokens": max_tokens,
            "stream": stream,
        }
        if temperature is not None:
            payload["temperature"] = temperature
        if top_p is not None:
            payload["top_p"] = top_p
        payload.update(kwargs)

        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())

        if stream:
            return self._create_stream(request_id, created_time, model, payload)
        else:
            return self._create_non_stream(request_id, created_time, model, payload)

    def _create_stream(
        self, request_id: str, created_time: int, model: str, payload: Dict[str, Any]
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            response = self._client.scraper.post(
                self._client.base_url,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=self._client.timeout
            )
            response.raise_for_status()

            prompt_tokens = 0
            completion_tokens = 0
            total_tokens = 0

            for line in response.iter_lines():
                if line:
                    decoded_line = line.decode('utf-8', errors='replace').strip()
                    if decoded_line.startswith("data: "):
                        json_str = decoded_line[6:]
                        if json_str == "[DONE]":
                            break
                        try:
                            data = json.loads(json_str)
                            choices = data.get('choices')
                            if not choices:
                                continue
                            choice_data = choices[0] if choices else {}
                            delta_data = choice_data.get('delta', {})
                            finish_reason = choice_data.get('finish_reason')

                            usage_data = data.get('usage', {})
                            if usage_data:
                                prompt_tokens = usage_data.get('prompt_tokens', prompt_tokens)
                                completion_tokens = usage_data.get('completion_tokens', completion_tokens)
                                total_tokens = usage_data.get('total_tokens', total_tokens)

                            delta = ChoiceDelta(
                                content=delta_data.get('content'),
                                role=delta_data.get('role'),
                                tool_calls=delta_data.get('tool_calls')
                            )
                            choice = Choice(
                                index=choice_data.get('index', 0),
                                delta=delta,
                                finish_reason=finish_reason,
                                logprobs=choice_data.get('logprobs')
                            )
                            chunk = ChatCompletionChunk(
                                id=request_id,
                                choices=[choice],
                                created=created_time,
                                model=model,
                                system_fingerprint=data.get('system_fingerprint')
                            )
                            if hasattr(chunk, "model_dump"):
                                chunk_dict = chunk.model_dump(exclude_none=True)
                            else:
                                chunk_dict = chunk.dict(exclude_none=True)
                            usage_dict = {
                                "prompt_tokens": prompt_tokens or 10,
                                "completion_tokens": completion_tokens or (len(delta_data.get('content', '')) if delta_data.get('content') else 0),
                                "total_tokens": total_tokens or (10 + (len(delta_data.get('content', '')) if delta_data.get('content') else 0)),
                                "estimated_cost": None
                            }
                            if delta_data.get('content'):
                                completion_tokens += 1
                                total_tokens = prompt_tokens + completion_tokens
                                usage_dict["completion_tokens"] = completion_tokens
                                usage_dict["total_tokens"] = total_tokens
                            chunk_dict["usage"] = usage_dict
                            yield chunk
                        except json.JSONDecodeError:
                            print(f"Warning: Could not decode JSON line: {json_str}")
                            continue
        except Exception as e:
            print(f"Error during ChutesAI stream request: {e}")
            raise IOError(f"ChutesAI request failed: {e}") from e
        except Exception as e:
            print(f"Error processing ChutesAI stream: {e}")
            raise

    def _create_non_stream(
        self, request_id: str, created_time: int, model: str, payload: Dict[str, Any]
    ) -> ChatCompletion:
        try:
            response = self._client.scraper.post(
                self._client.base_url,
                headers=self._client.headers,
                json=payload,
                timeout=self._client.timeout
            )
            response.raise_for_status()
            data = response.json()
            choices_data = data.get('choices', [])
            usage_data = data.get('usage', {})
            choices = []
            for choice_d in choices_data:
                message_d = choice_d.get('message', {})
                message = ChatCompletionMessage(
                    role=message_d.get('role', 'assistant'),
                    content=message_d.get('content', '')
                )
                choice = Choice(
                    index=choice_d.get('index', 0),
                    message=message,
                    finish_reason=choice_d.get('finish_reason', 'stop')
                )
                choices.append(choice)
            usage = CompletionUsage(
                prompt_tokens=usage_data.get('prompt_tokens', 0),
                completion_tokens=usage_data.get('completion_tokens', 0),
                total_tokens=usage_data.get('total_tokens', 0)
            )
            completion = ChatCompletion(
                id=request_id,
                choices=choices,
                created=created_time,
                model=data.get('model', model),
                usage=usage,
            )
            return completion
        except Exception as e:
            print(f"Error during ChutesAI non-stream request: {e}")
            raise IOError(f"ChutesAI request failed: {e}") from e
        except Exception as e:
            print(f"Error processing ChutesAI response: {e}")
            raise

class Chat(BaseChat):
    def __init__(self, client: 'ChutesAI'):
        self.completions = Completions(client)

class ChutesAI(OpenAICompatibleProvider):
    AVAILABLE_MODELS = [
        "deepseek-ai/DeepSeek-V3-0324",
        "deepseek-ai/DeepSeek-R1",
        "NousResearch/DeepHermes-3-Mistral-24B-Preview",
        "chutesai/Llama-4-Maverick-17B-128E-Instruct-FP8",
    ]
    def __init__(self, api_key: Optional[str] = None,):
        self.timeout = None  # Infinite timeout
        self.base_url = "https://llm.chutes.ai/v1/chat/completions"

        # Always generate a new API key, ignore any provided key
        print("[yellow]Generating new ChutesAI API key...[/]")
        self.api_key = generate_chutesai_api_key()

        if not self.api_key:
            print("[red]Failed to generate API key. Retrying...[/]")
            # Retry once more
            self.api_key = generate_chutesai_api_key()

        if not self.api_key:
            raise ValueError("Failed to generate ChutesAI API key after multiple attempts.")

        print(f"[green]Successfully generated API key: {self.api_key[:20]}...[/]")

        self.scraper = cloudscraper.create_scraper()
        self.headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json"
        }
        self.scraper.headers.update(self.headers)
        self.chat = Chat(self)

    @property
    def models(self) -> SimpleModelList:
        return SimpleModelList(type(self).AVAILABLE_MODELS)

if __name__ == "__main__":
    try:
        # Example usage - always use generated API key
        client = ChutesAI()

        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "What is the capital of France?"}
        ]

        print("[cyan]Making API request...[/]")
        response = client.chat.completions.create(
            model="deepseek-ai/DeepSeek-V3-0324",
            messages=messages,
            max_tokens=50,
            stream=True
        )
        for chunk in response:
            if isinstance(chunk, (bytes, str)):
                chunk_dict = {"chunk": chunk}
            elif hasattr(chunk, "model_dump") and callable(chunk.model_dump):
                chunk_dict = chunk.model_dump()  # type: ignore[call-arg]
            elif hasattr(chunk, "dict") and callable(chunk.dict):
                chunk_dict = chunk.dict()  # type: ignore[call-arg]
            else:
                chunk_dict = {"chunk": chunk}
            print(f"[green]Response Chunk:[/] {chunk_dict}")

    except Exception as e:
        print(f"[red]Error: {e}[/]")
        print("[yellow]If the issue persists, the ChutesAI service might be down or the API key generation method needs updating.[/]")



"""Lightweight proxy manager for streaming support and provider integration.

This module provides a simple, monkey-patchable proxy manager that rotates
through a pool of proxies. It works with both requests.Session and curl_cffi.

Usage:
    >>> from webscout.Extra.proxy_manager import ProxyManager
    >>> pm = ProxyManager(['http://proxy1:8080', 'http://proxy2:8080'])
    >>> pm.monkey_patch()  # Patch globally
    >>> # Now all providers use rotating proxies automatically
"""

import asyncio
import threading
import time
from itertools import cycle
from pathlib import Path
from typing import Iterator, Optional

import httpx
import requests
from curl_cffi.requests import Session as CurlSession
from litprinter import ic


class ProxyManager:
    """Manages and rotates proxies with background Urban VPN fetching support."""

    def __init__(
        self,
        proxies: Optional[list[str]] = None,
        filepath: Optional[str] = None,
        auto_fetch: bool = False,
        debug: bool = False,
    ) -> None:
        self._proxies: list[str] = self._load_initial(proxies, filepath)
        self._pool: Iterator[str] = cycle(self._proxies) if self._proxies else iter([])
        self._lock = threading.Lock()
        self.debug = debug
        self._headers = {
            "origin": "chrome-extension://eppiocemhmnlbhjplcgkofciiegomcon",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36",
        }
        if auto_fetch or not self._proxies:
            threading.Thread(target=self._background_refresh, daemon=True).start()

    def _load_initial(self, proxies: Optional[list[str]], filepath: Optional[str]) -> list[str]:
        if proxies:
            return [p.strip() for p in proxies if p.strip()]
        if filepath and (p := Path(filepath)).exists():
            return [line.strip() for line in p.read_text().splitlines() if line.strip()]
        return []

    async def _fetch_urban_proxies(self) -> list[str]:
        """Fetch proxies from Urban VPN API."""
        async with httpx.AsyncClient(headers=self._headers, timeout=10) as client:
            try:
                # 1. Anon Token
                r = await client.post(
                    "https://api-pro.falais.com/rest/v1/registrations/clientApps/URBAN_VPN_BROWSER_EXTENSION/users/anonymous",
                    json={"clientApp": {"name": "URBAN_VPN_BROWSER_EXTENSION", "browser": "CHROME"}},
                )
                r.raise_for_status()
                anon_response = r.json()
                if "value" not in anon_response:
                    if self.debug:
                        ic(f"Urban VPN anon response: {anon_response}")
                    return []
                anon_token = anon_response["value"]

                # 2. Access Token
                r = await client.post(
                    "https://api-pro.falais.com/rest/v1/security/tokens/accs",
                    headers={"authorization": f"Bearer {anon_token}"},
                    json={"type": "accs", "clientApp": {"name": "URBAN_VPN_BROWSER_EXTENSION"}},
                )
                r.raise_for_status()
                accs = r.json()
                if "value" not in accs:
                    if self.debug:
                        ic(f"Urban VPN accs response: {accs}")
                    return []
                token = accs["value"]

                # 3. Countries
                r = await client.get(
                    "https://stats.falais.com/api/rest/v2/entrypoints/countries",
                    headers={"authorization": f"Bearer {token}", "x-client-app": "URBAN_VPN_BROWSER_EXTENSION"},
                )
                r.raise_for_status()
                countries_response = r.json()
                if "countries" not in countries_response or "elements" not in countries_response["countries"]:
                    if self.debug:
                        ic(f"Urban VPN countries response: {countries_response}")
                    return []
                countries = countries_response["countries"]["elements"]

                # 4. Fetch credentials for a few random servers to keep it fast
                new_proxies = []
                for country in countries[:10]:  # Limit to 10 countries for speed
                    if "servers" not in country or not country["servers"]["elements"]:
                        continue
                    for server in country["servers"]["elements"][:1]:  # 1 server per country
                        if not server.get("signature"):
                            continue
                        r = await client.post(
                            "https://api-pro.falais.com/rest/v1/security/tokens/accs-proxy",
                            headers={"authorization": f"Bearer {token}"},
                            json={
                                "type": "accs-proxy",
                                "clientApp": {"name": "URBAN_VPN_BROWSER_EXTENSION"},
                                "signature": server["signature"],
                            },
                        )
                        r.raise_for_status()
                        cred_response = r.json()
                        if "value" not in cred_response:
                            if self.debug:
                                ic(f"Urban VPN cred response: {cred_response}")
                            continue
                        cred = cred_response["value"]
                        if "address" not in server or "primary" not in server["address"]:
                            continue
                        server_addr = server["address"]["primary"]
                        if "ip" not in server_addr or "port" not in server_addr:
                            continue
                        new_proxies.append(
                            f"http://{cred}:{cred}@{server_addr['ip']}:{server_addr['port']}"
                        )
                return new_proxies
            except Exception as e:
                if self.debug:
                    ic(f"Proxy fetch failed: {e}")
                return []

    def _background_refresh(self) -> None:
        """Background loop to refresh proxies."""
        while True:
            proxies = asyncio.run(self._fetch_urban_proxies())
            if proxies:
                with self._lock:
                    self._proxies = proxies
                    self._pool = cycle(self._proxies)
                if self.debug:
                    ic(f"Refreshed {len(proxies)} proxies background")
            time.sleep(1800)  # Refresh every 30 mins

    def get(self) -> dict[str, str]:
        """Get next proxy in rotation."""
        with self._lock:
            if not self._proxies:
                return {}
            proxy = next(self._pool)
            return {"http": proxy, "https": proxy}

    def patch(self) -> None:
        """Patch requests and curl_cffi to use this manager."""
        original_req_init = requests.Session.__init__
        original_curl_init = CurlSession.__init__
        pm = self

        def patched_req(self: requests.Session, *a, **k) -> None:
            original_req_init(self, *a, **k)
            if p := pm.get():
                self.proxies.update(p)

        def patched_curl(self: CurlSession, *a, **k) -> None:
            k["proxies"] = k.get("proxies") or pm.get()
            original_curl_init(self, *a, **k)

        requests.Session.__init__ = patched_req  # type: ignore
        CurlSession.__init__ = patched_curl  # type: ignore

    def monkey_patch(self) -> None:
        """Alias for patch() for backward compatibility."""
        self.patch()


__all__ = ["ProxyManager"]



"""
Brave Search AI Chat API Provider

This module provides a complete implementation of the Brave Search AI Chat API.
It supports both streaming and non-streaming responses using the real API endpoints.

Source: https://search.brave.com/ask
API Endpoint: https://search.brave.com/api/tap/v1

Features:
    - AI-generated responses based on web search
    - Streaming and non-streaming modes
    - Deep research capability
    - Proper error handling and session management

Generated by ReVibe.
Co-Authored-By: ReVibe <revibe@oevortex.dev>

STATUS: UNFINISHED
Issue: API endpoint returns 404 error on /new endpoint initialization
The reverse-engineered endpoints appear correct, but the actual API
may have changed or requires additional authentication/parameters.
Requires further investigation and testing against the live API.
"""

import asyncio
import base64
import json
import os
from typing import Any, AsyncIterator, Dict, Generator, List, Optional, Union, cast
from urllib.parse import quote

from curl_cffi.requests import AsyncSession

from webscout import exceptions
from webscout.AIbase import AISearch, SearchResponse


class BraveAI(AISearch):
    """
    A class to interact with Brave Search AI Chat API.

    BraveAI provides AI-generated responses based on web search results using the
    official Brave Search API. It supports both streaming and non-streaming responses.

    The API uses a symmetric_key for request authentication and requires various
    context parameters for proper operation.

    Basic Usage:
        >>> from webscout.Provider.AISEARCH import BraveAI
        >>> ai = BraveAI()
        >>> response = ai.search("What is Python?")
        >>> print(response)

    Streaming Usage:
        >>> ai = BraveAI()
        >>> for chunk in ai.search("Tell me about AI", stream=True):
        ...     print(chunk, end="", flush=True)

    Args:
        timeout (int, optional): Request timeout in seconds. Defaults to 30.
        proxies (dict, optional): Proxy configuration for requests. Defaults to None.
        language (str, optional): Language for search results. Defaults to "en".
        country (str, optional): Country code for localization. Defaults to "US".
        ui_lang (str, optional): UI language code. Defaults to "en-us".
        geoloc (str, optional): Geolocation in format "latitude,longitude".
                               Defaults to None (will use default).
        safesearch (str, optional): Safe search level. Defaults to "moderate".
                                   Options: "off", "moderate", "strict".
        units (str, optional): Units of measurement. Defaults to "metric".
        use_location (bool, optional): Whether to use geolocation. Defaults to True.
    """

    BASE_URL = "https://search.brave.com/api/tap/v1"

    def __init__(
        self,
        timeout: int = 30,
        proxies: Optional[dict] = None,
        language: str = "en",
        country: str = "US",
        ui_lang: str = "en-us",
        geoloc: Optional[str] = None,
        safesearch: str = "moderate",
        units: str = "metric",
        use_location: bool = True,
    ):
        """Initialize BraveAI with configuration parameters."""
        self.timeout = timeout
        self.proxies = proxies or {}
        self.language = language
        self.country = country.lower()
        self.ui_lang = ui_lang
        self.geoloc = geoloc
        self.safesearch = safesearch
        self.units = units
        self.use_location = "1" if use_location else "0"
        self.last_response = {}
        self.chat_id = None
        self.symmetric_key = None
        self.headers = {
            "accept": "application/json",
            "accept-language": "en-US,en;q=0.9",
            "user-agent": (
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36"
            ),
            "sec-ch-ua": (
                '"Not(A:Brand";v="8", "Chromium";v="144", "Google Chrome";v="144"'
            ),
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "referer": "https://search.brave.com/ask",
        }

    def _generate_symmetric_key(self) -> str:
        """
        Generate a symmetric encryption key for API authentication.

        Creates a 256-bit AES-GCM key in JWK (JSON Web Key) format, base64 encoded.

        Returns:
            str: Base64-encoded JWK representation of the encryption key.
        """
        key = os.urandom(32)
        k = base64.urlsafe_b64encode(key).decode().rstrip("=")
        jwk = {
            "alg": "A256GCM",
            "ext": True,
            "k": k,
            "key_ops": ["encrypt", "decrypt"],
            "kty": "oct",
        }
        return base64.b64encode(
            json.dumps(jwk, separators=(",", ":")).encode()
        ).decode()

    def _build_context_params(
        self, enable_research: bool = False
    ) -> Dict[str, Union[str, int]]:
        """
        Build common context parameters for API requests.

        These parameters are required by the Brave Search API for proper
        context and localization of responses.

        Args:
            enable_research (bool): Whether to enable deep research mode.

        Returns:
            dict: Dictionary of context parameters.
        """
        params = {
            "language": self.language,
            "country": self.country,
            "ui_lang": self.ui_lang,
            "safesearch": self.safesearch,
            "force_safesearch": "0",
            "units_of_measurement": self.units,
            "use_location": self.use_location,
        }

        if self.geoloc:
            params["geoloc"] = self.geoloc

        return params

    def search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        enable_research: bool = False,
        **kwargs: Any,
    ) -> Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]]:
        """
        Search using the Brave Search AI and get AI-generated responses.

        This method sends a search query to Brave AI and returns the AI-generated
        response. It supports both streaming and non-streaming modes, as well as
        raw response format.

        Args:
            prompt (str): The search query or prompt to send to the API.
            stream (bool, optional): If True, yields response chunks as they arrive.
                                   If False, returns complete response. Defaults to False.
            raw (bool, optional): If True, returns raw response dictionaries with
                                'text' key. If False, returns Response objects.
                                Defaults to False.
            enable_research (bool, optional): If True, enables deep research mode
                                             for more comprehensive responses.
                                             Defaults to False.
            **kwargs: Additional keyword arguments (for compatibility).

        Returns:
            Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]]:
                - If stream=False: Returns complete response as SearchResponse object
                - If stream=True: Yields response chunks as either Dict or SearchResponse

        Raises:
            APIConnectionError: If the API request fails.
            FailedToGenerateResponseError: If response parsing fails.

        Examples:
            Basic search:
            >>> ai = BraveAI()
            >>> response = ai.search("What is Python?")
            >>> print(response)

            Research-enabled search:
            >>> response = ai.search("What is Python?", enable_research=True)
            >>> print(response)

            Streaming response:
            >>> for chunk in ai.search("Tell me about AI", stream=True):
            ...     print(chunk, end="", flush=True)
        """
        if not stream:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                result = loop.run_until_complete(
                    self._async_search(prompt, False, raw, enable_research)
                )
                return cast(
                    Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]],
                    result,
                )
            finally:
                loop.close()

        buffer = ""

        def sync_generator():
            nonlocal buffer
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                async_gen = loop.run_until_complete(
                    self._async_search(prompt, True, raw, enable_research)
                )
                if hasattr(async_gen, "__anext__"):
                    async_iterator = cast(AsyncIterator, async_gen)
                    while True:
                        try:
                            chunk = loop.run_until_complete(async_iterator.__anext__())
                            if isinstance(chunk, dict) and "text" in chunk:
                                buffer += chunk["text"]
                            elif isinstance(chunk, SearchResponse):
                                buffer += chunk.text
                            yield chunk
                        except StopAsyncIteration:
                            break
                elif isinstance(async_gen, SearchResponse):
                    yield async_gen
            finally:
                self.last_response = {"text": buffer}
                loop.close()

        return cast(
            Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]],
            sync_generator(),
        )

    async def _async_search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        enable_research: bool = False,
    ) -> Union[
        SearchResponse,
        str,
        AsyncIterator[Union[Dict[str, str], SearchResponse]],
    ]:
        """
        Internal async implementation of the search method.

        Handles the actual API communication, session management, and
        response processing.

        Args:
            prompt (str): The search query.
            stream (bool): Whether to stream the response.
            raw (bool): Whether to return raw response format.
            enable_research (bool): Whether to enable deep research.

        Returns:
            Union[SearchResponse, str, AsyncIterator]: Response data.

        Raises:
            FailedToGenerateResponseError: If API communication fails.
        """
        session = AsyncSession()
        try:
            session.headers.update(self.headers)
            if self.proxies:
                session.proxies.update(self.proxies)

            # Initialize new chat session if needed
            if not self.chat_id or not self.symmetric_key:
                self.symmetric_key = self._generate_symmetric_key()

                # Build context parameters for new session
                params = self._build_context_params(enable_research)
                params["symmetric_key"] = self.symmetric_key
                params["source"] = "home"
                params["enable_research"] = "true" if enable_research else "false"

                try:
                    response = await session.get(
                        f"{self.BASE_URL}/new",
                        params=params,
                        timeout=self.timeout,
                        impersonate="chrome110",
                    )
                    response.raise_for_status()
                except Exception as e:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Failed to create new chat session: {str(e)}"
                    )

                try:
                    data = response.json()
                    self.chat_id = data.get("id")
                    if not self.chat_id:
                        raise exceptions.FailedToGenerateResponseError(
                            "No conversation ID returned from API"
                        )
                except json.JSONDecodeError as e:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Invalid JSON response: {str(e)}"
                    )

            # Build stream request parameters
            params = self._build_context_params(enable_research)
            params["id"] = self.chat_id
            params["query"] = prompt
            params["symmetric_key"] = self.symmetric_key
            params["enable_followups"] = "true" if enable_research else "false"

            # Set proper referer for the request
            referer = f"https://search.brave.com/ask?q={quote(prompt)}&conversation={self.chat_id}"
            headers = {"referer": referer}

            if not stream:
                # Non-streaming mode: collect entire response
                full_text = ""
                response = await session.get(
                    f"{self.BASE_URL}/stream",
                    params=params,
                    headers=headers,
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome110",
                )
                response.raise_for_status()

                async for line in response.aiter_lines():
                    if not line:
                        continue
                    try:
                        j = json.loads(line)
                        if j.get("type") == "text_delta":
                            full_text += j.get("delta", "")
                    except json.JSONDecodeError:
                        continue

                self.last_response = SearchResponse(full_text)
                return full_text if raw else self.last_response

            # Streaming mode: yield chunks as they arrive
            async def process_stream():
                buffer = ""
                try:
                    response = await session.get(
                        f"{self.BASE_URL}/stream",
                        params=params,
                        headers=headers,
                        stream=True,
                        timeout=self.timeout,
                        impersonate="chrome110",
                    )
                    response.raise_for_status()

                    async for line in response.aiter_lines():
                        if not line:
                            continue
                        try:
                            j = json.loads(line)
                            t = j.get("type")
                            if t == "text_delta":
                                delta = j.get("delta", "")
                                buffer += delta
                                if raw:
                                    yield j
                                else:
                                    yield SearchResponse(delta)
                        except json.JSONDecodeError:
                            continue
                    self.last_response = SearchResponse(buffer)
                finally:
                    await session.close()

            return process_stream()
        except Exception:
            if not stream:
                await session.close()
            raise


if __name__ == "__main__":
    # Example usage
    ai = BraveAI()

    # Non-streaming test
    print("=== Non-streaming Response ===")
    res = ai.search("What is Python?", stream=False)
    print(f"Result: {res}\n")

    # Streaming test
    print("=== Streaming Response ===")
    res = ai.search("Explain machine learning", stream=True)
    from collections.abc import Iterable

    if isinstance(res, Iterable):
        for chunk in res:
            print(chunk, end="", flush=True)
    else:
        print(res)

    print("\n=== Done ===")



import re
from typing import Any, Dict, Generator, List, Optional, Union, cast

import requests

from webscout import exceptions
from webscout.AIbase import AISearch, SearchResponse
from webscout.litagent import LitAgent
from webscout.sanitize import sanitize_stream


class webpilotai(AISearch):
    """A class to interact with the webpilotai (WebPilot) AI search API.

    webpilotai provides a web-based comprehensive search SearchResponse interface that returns AI-generated
    SearchResponses with source references and related questions. It supports both streaming and
    non-streaming SearchResponses.

    Basic Usage:
        >>> from webscout import webpilotai
        >>> ai = webpilotai()
        >>> # Non-streaming example
        >>> response = ai.search("What is Python?")
        >>> print(response)
        Python is a high-level programming language...

        >>> # Streaming example
        >>> for chunk in ai.search("Tell me about AI", stream=True):
        ...     print(chunk, end="", flush=True)
        Artificial Intelligence is...

        >>> # Raw SearchResponse format
        >>> for chunk in ai.search("Hello", stream=True, raw=True):
        ...     print(chunk)
        {'text': 'Hello'}
        {'text': ' there!'}

    Args:
        timeout (int, optional): Request timeout in seconds. Defaults to 90.
        proxies (dict, optional): Proxy configuration for requests. Defaults to None.
    """

    def __init__(
        self,
        timeout: int = 90,
        proxies: Optional[dict] = None,
    ):
        """Initialize the webpilotai API client.

        Args:
            timeout (int, optional): Request timeout in seconds. Defaults to 90.
            proxies (dict, optional): Proxy configuration for requests. Defaults to None.
        """
        self.session = requests.Session()
        self.api_endpoint = "https://api.webpilotai.com/rupee/v1/search"
        self.timeout = timeout
        self.last_response = {}

        # The 'Bearer null' is part of the API's expected headers
        self.headers = {
            'Accept': 'application/json, text/plain, */*, text/event-stream',
            'Content-Type': 'application/json;charset=UTF-8',
            'Authorization': 'Bearer null',
            'Origin': 'https://www.webpilot.ai',
            'Referer': 'https://www.webpilot.ai/',
            'User-Agent': LitAgent().random(),
        }

        self.session.headers.update(self.headers)
        self.proxies = proxies

    def search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None], List[Any], Dict[str, Any], str]:
        """Search using the webpilotai API and get AI-generated SearchResponses.

        This method sends a search query to webpilotai and returns the AI-generated SearchResponse.
        It supports both streaming and non-streaming modes, as well as raw SearchResponse format.

        Args:
            prompt (str): The search query or prompt to send to the API.
            stream (bool, optional): If True, yields SearchResponse chunks as they arrive.
                                   If False, returns complete SearchResponse. Defaults to False.
            raw (bool, optional): If True, returns raw SearchResponse dictionaries with 'text' key.
                                If False, returns SearchResponse objects that convert to text automatically.
                                Defaults to False.

        Returns:
            Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]]:
                - If stream=False: Returns complete SearchResponse as SearchResponse object
                - If stream=True: Yields SearchResponse chunks as either Dict or SearchResponse objects

        Raises:
            APIConnectionError: If the API request fails
        """
        payload = {
            "q": prompt,
            "threadId": ""  # Empty for new search
        }

        # We'll use regex-based extraction for WebPilot events. The sanitizer will:
        #  - skip lines that are not "type":"data" events using skip_regexes
        #  - extract the "content" value using a regex that handles escaped quotes
        # The regex used below for extract_regexes is: r'"content"\s*:\s*"((?:\\.|[^"\\])*)"'
        # which captures the content string while allowing escaped characters such as \".

        def for_stream():

            try:
                with self.session.post(
                    self.api_endpoint,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    proxies=self.proxies
                ) as response:
                    if not response.ok:
                        raise exceptions.APIConnectionError(
                            f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                        )

                    processed_chunks = sanitize_stream(
                        data=response.iter_content(chunk_size=1024),
                        to_json=True,
                        # Extract content from parsed JSON payload similar to Monica
                        content_extractor=lambda chunk: (
                            ((chunk.get("data") or {}).get("content") if isinstance(chunk, dict) else None)
                            or ((chunk.get("data") or {}).get("text") if isinstance(chunk, dict) else None)
                            or ((chunk.get("data") or {}).get("delta", {}).get("content") if isinstance(chunk, dict) else None)
                        ),
                        skip_markers=["event:message"],
                        yield_raw_on_error=False,
                        encoding='utf-8',
                        encoding_errors='replace',
                        line_delimiter="\n",
                        raw=raw,
                        output_formatter=None if raw else lambda x: SearchResponse(x) if isinstance(x, str) else x,
                    )

                    for chunk in processed_chunks:
                        yield chunk

            except requests.exceptions.Timeout:
                raise exceptions.APIConnectionError("Request timed out")
            except requests.exceptions.RequestException as e:
                raise exceptions.APIConnectionError(f"Request failed: {e}")

        def for_non_stream():
            try:
                with self.session.post(
                    self.api_endpoint,
                    json=payload,
                    stream=False,
                    timeout=self.timeout,
                    proxies=self.proxies,
                ) as response:
                    if not response.ok:
                        raise exceptions.APIConnectionError(
                            f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                        )

                    if raw:
                        return response.text

                    # Process full response payload using sanitize_stream similar to streaming path
                    processed_chunks = sanitize_stream(
                        data=response.content,
                        intro_value="",
                        to_json=True,
                        strip_chars=None,
                        start_marker=None,
                        end_marker=None,
                        content_extractor=lambda chunk: (
                            ((chunk.get("data") or {}).get("content") if isinstance(chunk, dict) else None)
                            or ((chunk.get("data") or {}).get("text") if isinstance(chunk, dict) else None)
                            or ((chunk.get("data") or {}).get("delta", {}).get("content") if isinstance(chunk, dict) else None)
                        ),
                        skip_markers=["event:message"],
                        yield_raw_on_error=False,
                        encoding='utf-8',
                        encoding_errors='replace',
                        buffer_size=8192,
                        output_formatter=lambda x: SearchResponse(x) if isinstance(x, str) else x,
                    )

                    full_response = ""
                    for content_chunk in processed_chunks:
                        if content_chunk is not None and isinstance(content_chunk, str):
                            full_response += content_chunk

                    formatted_response = self.format_SearchResponse(full_response)
                    self.last_response = SearchResponse(formatted_response)
                    return self.last_response

            except requests.exceptions.RequestException as e:
                raise exceptions.APIConnectionError(f"Request failed: {e}")

        if stream:
            return for_stream()
        else:
            return for_non_stream()

    @staticmethod
    def format_SearchResponse(text: str) -> str:
        """Format the SearchResponse text for better readability.

        Args:
            text (str): The raw SearchResponse text

        Returns:
            str: Formatted text with improved structure
        """
        # Clean up formatting
        # Remove excessive newlines
        clean_text = re.sub(r'\n{3,}', '\n\n', text)

        # Ensure consistent spacing around sections
        clean_text = re.sub(r'([.!?])\s*\n\s*([A-Z])', r'\1\n\n\2', clean_text)

        # Clean up any leftover HTML or markdown artifacts
        clean_text = re.sub(r'<[^>]*>', '', clean_text)

        # Remove trailing whitespace on each line
        clean_text = '\n'.join(line.rstrip() for line in clean_text.split('\n'))

        return clean_text.strip()


if __name__ == "__main__":
    from rich import print

    ai = webpilotai()
    resp = ai.search("webscout python package details", stream=True, raw=False)
    if hasattr(resp, "__iter__") and not isinstance(resp, (str, SearchResponse)):
        for chunk in resp:
            print(chunk, end="", flush=True)
    else:
        print(resp)



##################################################################################
##  Deepgram TTS Provider                                                      ##
##################################################################################
import pathlib
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Optional, Union, cast

import requests
from litprinter import ic

from webscout import exceptions
from webscout.litagent import LitAgent

try:
    from . import utils
    from .base import BaseTTSProvider
except ImportError:
    # Handle direct execution
    import os
    import sys

    sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "..", ".."))
    from webscout.Provider.TTS import utils
    from webscout.Provider.TTS.base import BaseTTSProvider


class DeepgramTTS(BaseTTSProvider):
    """
    Text-to-speech provider using the Deepgram Aura-2 API.

    This provider follows the OpenAI TTS API structure with support for:
    - Aura-2 next-gen voices
    - Low-latency real-time performance
    - Multiple output formats
    - Concurrent generation for long texts
    """

    required_auth = False

    # Request headers
    headers: dict[str, str] = {
        "Accept": "*/*",
        "Accept-Language": "en-US,en;q=0.9",
        "Content-Type": "application/json",
        "Origin": "https://deepgram.com",
        "Referer": "https://deepgram.com/ai-voice-generator",
        "User-Agent": LitAgent().random(),
    }

    # Supported Aura-2 voices
    SUPPORTED_MODELS = ["aura-2"]

    SUPPORTED_VOICES = [
        "thalia",
        "odysseus",
        "harmonia",
        "theia",
        "electra",
        "arcas",
        "amalthea",
        "helena",
        "hyperion",
        "apollo",
        "luna",
        # Legacy Aura-1 voices (if still supported by the endpoint)
        "asteria",
        "luna",
        "stella",
        "athena",
        "hera",
        "zeus",
        "orpheus",
        "arcas",
        "perseus",
        "angus",
        "orion",
        "helios",
    ]

    # Voice mapping for Deepgram API compatibility
    voice_mapping = {
        # Aura-2
        "thalia": "aura-2-thalia-en",
        "odysseus": "aura-2-odysseus-en",
        "harmonia": "aura-2-harmonia-en",
        "theia": "aura-2-theia-en",
        "electra": "aura-2-electra-en",
        "arcas": "aura-2-arcas-en",
        "amalthea": "aura-2-amalthea-en",
        "helena": "aura-2-helena-en",
        "hyperion": "aura-2-hyperion-en",
        "apollo": "aura-2-apollo-en",
        "luna": "aura-2-luna-en",
        # Aura-1 (Backward compatibility)
        "asteria": "aura-asteria-en",
        "stella": "aura-stella-en",
        "athena": "aura-athena-en",
        "hera": "aura-hera-en",
        "zeus": "aura-zeus-en",
        "orpheus": "aura-orpheus-en",
        "perseus": "aura-perseus-en",
        "angus": "aura-angus-en",
        "orion": "aura-orion-en",
        "helios": "aura-helios-en",
    }

    def __init__(self, timeout: int = 30, proxies: Optional[dict] = None):
        """
        Initialize the Deepgram TTS client.

        Args:
            timeout (int): Request timeout in seconds
            proxies (dict): Proxy configuration
        """
        super().__init__()
        self.api_url = "https://deepgram.com/api/ttsAudioGeneration"
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)
        self.timeout = timeout
        self.default_voice = "thalia"

    def tts(self, text: str, voice: Optional[str] = None, verbose: bool = False, **kwargs) -> str:
        """
        Convert text to speech using Deepgram Aura-2 API.

        Args:
            text (str): The text to convert to speech
            voice (str): The voice to use (thalia, odysseus, etc.)
            verbose (bool): Whether to print debug information
            **kwargs: Additional parameters

        Returns:
            str: Path to the generated audio file
        """
        # Extract parameters from kwargs with defaults
        voice = voice or kwargs.get("voice", "thalia")
        response_format = kwargs.get("response_format", "mp3")
        verbose = verbose if verbose is not None else kwargs.get("verbose", True)

        if not text:
            raise ValueError("Input text must be a non-empty string")

        # Map voice to Deepgram API format
        voice_id = self.voice_mapping.get(voice.lower(), f"aura-2-{voice.lower()}-en")

        # Create temporary file
        file_extension = f".{response_format}"
        filename = pathlib.Path(
            tempfile.NamedTemporaryFile(suffix=file_extension, dir=self.temp_dir, delete=False).name
        )
        sentences = utils.split_sentences(text)
        if verbose:
            ic.configureOutput(prefix="DEBUG| ")
            ic(f"DeepgramTTS: Processing {len(sentences)} chunks")
            ic.configureOutput(prefix="DEBUG| ")
            ic(f"Voice: {voice} -> {voice_id}")

        def generate_audio_for_chunk(part_text: str, part_number: int):
            max_retries = 3
            for attempt in range(max_retries):
                try:
                    payload = {
                        "text": part_text,
                        "model": voice_id,
                        "demoType": "voice-generator",
                        "params": "tag=landingpage-aivoicegenerator",
                    }
                    response = self.session.post(self.api_url, json=payload, timeout=self.timeout)
                    response.raise_for_status()

                    if response.content:
                        if verbose:
                            ic.configureOutput(prefix="DEBUG| ")
                            ic(f"Chunk {part_number} processed successfully")
                        return part_number, response.content

                except requests.RequestException as e:
                    if verbose:
                        ic.configureOutput(prefix="WARNING| ")
                        ic(
                            f"Error processing chunk {part_number}: {e}. Retrying {attempt + 1}/{max_retries}"
                        )
                    time.sleep(1)

            raise exceptions.FailedToGenerateResponseError(
                f"Failed to generate audio for chunk {part_number}"
            )

        try:
            with ThreadPoolExecutor(max_workers=3) as executor:
                futures = {
                    executor.submit(generate_audio_for_chunk, sentence.strip(), i): i
                    for i, sentence in enumerate(sentences)
                }

                audio_chunks = {}
                for future in as_completed(futures):
                    part_num, data = future.result()
                    audio_chunks[part_num] = data

                with open(filename, "wb") as f:
                    for i in sorted(audio_chunks.keys()):
                        f.write(audio_chunks[i])

                if verbose:
                    ic.configureOutput(prefix="INFO| ")
                    ic(f"Audio saved to {filename}")

                return str(filename)

        except Exception as e:
            raise exceptions.FailedToGenerateResponseError(f"Deepgram TTS failed: {e}")

    def create_speech(
        self,
        input_text: str,
        model: Optional[str] = None,
        voice: Optional[str] = None,
        response_format: Optional[str] = None,
        instructions: Optional[str] = None,
        verbose: bool = False,
    ) -> str:
        return self.tts(
            text=input_text,
            model=model or "aura-2",
            voice=voice or "thalia",
            response_format=response_format or "mp3",
            instructions=instructions,
            verbose=verbose,
        )

    def with_streaming_response(self):
        return StreamingResponseContextManager(self)


class StreamingResponseContextManager:
    def __init__(self, tts_provider: DeepgramTTS):
        self.tts_provider = tts_provider

    def create(self, **kwargs):
        audio_file = self.tts_provider.create_speech(**kwargs)
        return StreamingResponse(audio_file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class StreamingResponse:
    def __init__(self, audio_file: str):
        self.audio_file = audio_file

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def stream_to_file(self, file_path: str):
        import shutil

        shutil.copy2(self.audio_file, file_path)

    def iter_bytes(self, chunk_size: int = 1024):
        with open(self.audio_file, "rb") as f:
            while chunk := f.read(chunk_size):
                yield chunk


if __name__ == "__main__":
    dg = DeepgramTTS()
    try:
        path = dg.tts(
            "Deepgram Aura-2 test successful. Generating high quality speech.",
            voice="thalia",
            verbose=True,
        )
        print(f"Saved to: {path}")
    except Exception as e:
        print(f"Error: {e}")



"""
Base class for TTS providers with OpenAI-compatible functionality.
"""

import asyncio
import functools
import os
import tempfile
from pathlib import Path
from typing import Any, Generator, List, Optional, Union, cast

from litprinter import ic

from webscout.AIbase import TTSProvider


class BaseTTSProvider(TTSProvider):
    """
    Base class for TTS providers with OpenAI-compatible functionality.

    This class implements common methods and follows OpenAI TTS API patterns
    for speech generation, streaming, and audio handling.
    """

    required_auth = False

    # Supported models (can be overridden by subclasses)
    SUPPORTED_MODELS = [
        "gpt-4o-mini-tts",  # Latest intelligent realtime model
        "tts-1",  # Lower latency model
        "tts-1-hd",  # Higher quality model
    ]

    # Supported voices (can be overridden by subclasses)
    SUPPORTED_VOICES = [
        "alloy",
        "ash",
        "ballad",
        "coral",
        "echo",
        "fable",
        "nova",
        "onyx",
        "sage",
        "shimmer",
    ]

    # Supported output formats
    SUPPORTED_FORMATS = [
        "mp3",  # Default format
        "opus",  # Internet streaming, low latency
        "aac",  # Digital audio compression
        "flac",  # Lossless compression
        "wav",  # Uncompressed, low latency
        "pcm",  # Raw samples, 24kHz 16-bit
    ]

    def __init__(self):
        """Initialize the base TTS provider."""
        self.temp_dir = tempfile.mkdtemp(prefix="webscout_tts_")
        self.default_model = "gpt-4o-mini-tts"
        self.default_voice = "coral"
        self.default_format = "mp3"

    def validate_model(self, model: str) -> str:
        """
        Validate and return the model name.

        Args:
            model (str): Model name to validate

        Returns:
            str: Validated model name

        Raises:
            ValueError: If model is not supported
        """
        # If provider doesn't support models, return the model as-is
        if self.SUPPORTED_MODELS is None:
            return model

        if model not in self.SUPPORTED_MODELS:
            raise ValueError(
                f"Model '{model}' not supported. Available models: {', '.join(self.SUPPORTED_MODELS)}"
            )
        return model

    def validate_voice(self, voice: str) -> str:
        """
        Validate and return the voice name.

        Args:
            voice (str): Voice name to validate

        Returns:
            str: Validated voice name

        Raises:
            ValueError: If voice is not supported
        """
        if voice not in self.SUPPORTED_VOICES:
            raise ValueError(
                f"Voice '{voice}' not supported. Available voices: {', '.join(self.SUPPORTED_VOICES)}"
            )
        return voice

    def validate_format(self, response_format: str) -> str:
        """
        Validate and return the response format.

        Args:
            response_format (str): Response format to validate

        Returns:
            str: Validated response format

        Raises:
            ValueError: If format is not supported
        """
        if response_format not in self.SUPPORTED_FORMATS:
            raise ValueError(
                f"Format '{response_format}' not supported. Available formats: {', '.join(self.SUPPORTED_FORMATS)}"
            )
        return response_format

    def save_audio(
        self, audio_file: str, destination: Optional[str] = None, verbose: bool = False
    ) -> str:
        """
        Save audio to a specific destination.

        Args:
            audio_file (str): Path to the source audio file
            destination (str, optional): Destination path. Defaults to current directory with timestamp.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Returns:
            str: Path to the saved audio file

        Raises:
            FileNotFoundError: If the audio file doesn't exist
        """
        import shutil
        import time

        source_path = Path(audio_file)

        if not source_path.exists():
            raise FileNotFoundError(f"Audio file not found: {audio_file}")

        if destination is None:
            # Create a default destination with timestamp in current directory
            timestamp = int(time.time())
            destination = os.path.join(os.getcwd(), f"speech_{timestamp}{source_path.suffix}")

        # Ensure the destination directory exists
        os.makedirs(os.path.dirname(os.path.abspath(destination)), exist_ok=True)

        # Copy the file
        shutil.copy2(source_path, destination)

        if verbose:
            ic.configureOutput(prefix="DEBUG| ")
            ic(f"Audio saved to {destination}")

        return destination

    def create_speech(
        self,
        input_text: str,
        model: Optional[str] = None,
        voice: Optional[str] = None,
        response_format: Optional[str] = None,
        instructions: Optional[str] = None,
        verbose: bool = False,
    ) -> str:
        """
        Create speech from input text (OpenAI-compatible interface).

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format (mp3, opus, aac, flac, wav, pcm)
            instructions (str, optional): Voice instructions for controlling speech aspects
            verbose (bool, optional): Whether to print debug information

        Returns:
            str: Path to the generated audio file
        """
        # Use defaults if not provided
        model = model or self.default_model
        voice = voice or self.default_voice
        response_format = response_format or self.default_format

        # Validate parameters
        self.validate_model(model)
        self.validate_voice(voice)
        self.validate_format(response_format)

        # Call the provider-specific TTS implementation
        return self.tts(
            text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose,
        )

    def stream_audio(
        self,
        text: str,
        model: Optional[str] = None,
        voice: Optional[str] = None,
        response_format: Optional[str] = None,
        instructions: Optional[str] = None,
        chunk_size: int = 1024,
        verbose: bool = False,
    ) -> Generator[bytes, None, None]:
        """
        Stream audio in chunks with OpenAI-compatible parameters.

        Args:
            text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format
            instructions (str, optional): Voice instructions
            chunk_size (int, optional): Size of audio chunks to yield. Defaults to 1024.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Yields:
            Generator[bytes, None, None]: Audio data chunks
        """
        # Generate the audio file using create_speech
        audio_file = self.create_speech(
            input_text=text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose,
        )

        # Stream the file in chunks
        with open(audio_file, "rb") as f:
            while chunk := f.read(chunk_size):
                yield chunk

    def tts(self, text: str, voice: Optional[str] = None, verbose: bool = False, **kwargs) -> str:
        """
        Abstract method for text-to-speech conversion.
        Must be implemented by subclasses.

        Args:
            text (str): The text to convert to speech
            **kwargs: Additional provider-specific parameters

        Returns:
            str: Path to the generated audio file

        Raises:
            NotImplementedError: If not implemented by subclass
        """
        raise NotImplementedError("Subclasses must implement the tts method")


class AsyncBaseTTSProvider:
    """
    Base class for async TTS providers with OpenAI-compatible functionality.

    This class implements common async methods following OpenAI TTS API patterns
    for speech generation, streaming, and audio handling.
    """

    required_auth = False

    # Supported models (can be overridden by subclasses)
    SUPPORTED_MODELS = [
        "gpt-4o-mini-tts",  # Latest intelligent realtime model
        "tts-1",  # Lower latency model
        "tts-1-hd",  # Higher quality model
    ]

    # Supported voices (can be overridden by subclasses)
    SUPPORTED_VOICES = [
        "alloy",
        "ash",
        "ballad",
        "coral",
        "echo",
        "fable",
        "nova",
        "onyx",
        "sage",
        "shimmer",
    ]

    # Supported output formats
    SUPPORTED_FORMATS = [
        "mp3",  # Default format
        "opus",  # Internet streaming, low latency
        "aac",  # Digital audio compression
        "flac",  # Lossless compression
        "wav",  # Uncompressed, low latency
        "pcm",  # Raw samples, 24kHz 16-bit
    ]

    def __init__(self):
        """Initialize the async base TTS provider."""
        self.temp_dir = tempfile.mkdtemp(prefix="webscout_tts_")
        self.default_model = "gpt-4o-mini-tts"
        self.default_voice = "coral"
        self.default_format = "mp3"

    async def validate_model(self, model: str) -> str:
        """
        Validate and return the model name.

        Args:
            model (str): Model name to validate

        Returns:
            str: Validated model name

        Raises:
            ValueError: If model is not supported
        """
        # If provider doesn't support models, return the model as-is
        if self.SUPPORTED_MODELS is None:
            return model

        if model not in self.SUPPORTED_MODELS:
            raise ValueError(
                f"Model '{model}' not supported. Available models: {', '.join(self.SUPPORTED_MODELS)}"
            )
        return model

    async def validate_voice(self, voice: str) -> str:
        """
        Validate and return the voice name.

        Args:
            voice (str): Voice name to validate

        Returns:
            str: Validated voice name

        Raises:
            ValueError: If voice is not supported
        """
        if voice not in self.SUPPORTED_VOICES:
            raise ValueError(
                f"Voice '{voice}' not supported. Available voices: {', '.join(self.SUPPORTED_VOICES)}"
            )
        return voice

    async def validate_format(self, response_format: str) -> str:
        """
        Validate and return the response format.

        Args:
            response_format (str): Response format to validate

        Returns:
            str: Validated response format

        Raises:
            ValueError: If format is not supported
        """
        if response_format not in self.SUPPORTED_FORMATS:
            raise ValueError(
                f"Format '{response_format}' not supported. Available formats: {', '.join(self.SUPPORTED_FORMATS)}"
            )
        return response_format

    async def save_audio(
        self, audio_file: str, destination: Optional[str] = None, verbose: bool = False
    ) -> str:
        """
        Save audio to a specific destination asynchronously.

        Args:
            audio_file (str): Path to the source audio file
            destination (str, optional): Destination path. Defaults to current directory with timestamp.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Returns:
            str: Path to the saved audio file

        Raises:
            FileNotFoundError: If the audio file doesn't exist
        """
        import asyncio
        import shutil
        import time

        source_path = Path(audio_file)

        if not source_path.exists():
            raise FileNotFoundError(f"Audio file not found: {audio_file}")

        if destination is None:
            # Create a default destination with timestamp in current directory
            timestamp = int(time.time())
            destination = os.path.join(os.getcwd(), f"speech_{timestamp}{source_path.suffix}")

        # Ensure the destination directory exists
        os.makedirs(os.path.dirname(os.path.abspath(destination)), exist_ok=True)

        # Copy the file using asyncio to avoid blocking
        await asyncio.to_thread(functools.partial(shutil.copy2, str(source_path), destination))

        if verbose:
            ic.configureOutput(prefix="DEBUG| ")
            ic(f"Audio saved to {destination}")

        return destination

    async def create_speech(
        self,
        input_text: str,
        model: Optional[str] = None,
        voice: Optional[str] = None,
        response_format: Optional[str] = None,
        instructions: Optional[str] = None,
        verbose: bool = False,
    ) -> str:
        """
        Create speech from input text asynchronously (OpenAI-compatible interface).

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format (mp3, opus, aac, flac, wav, pcm)
            instructions (str, optional): Voice instructions for controlling speech aspects
            verbose (bool, optional): Whether to print debug information

        Returns:
            str: Path to the generated audio file
        """
        # Use defaults if not provided
        model = model or self.default_model
        voice = voice or self.default_voice
        response_format = response_format or self.default_format

        # Validate parameters
        await self.validate_model(model)
        await self.validate_voice(voice)
        await self.validate_format(response_format)

        # Call the provider-specific TTS implementation
        return await self.tts(
            text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose,
        )

    async def stream_audio(
        self,
        input_text: str,
        model: Optional[str] = None,
        voice: Optional[str] = None,
        response_format: Optional[str] = None,
        instructions: Optional[str] = None,
        chunk_size: int = 1024,
        verbose: bool = False,
    ):
        """
        Stream audio in chunks asynchronously with OpenAI-compatible parameters.

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format
            instructions (str, optional): Voice instructions
            chunk_size (int, optional): Size of audio chunks to yield. Defaults to 1024.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Yields:
            AsyncGenerator[bytes, None]: Audio data chunks
        """
        # Generate the audio file using create_speech
        audio_file = await self.create_speech(
            input_text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose,
        )

        # Stream the file in chunks using thread pool
        def read_file_chunks(file_path, chunk_size):
            with open(file_path, "rb") as f:
                while chunk := f.read(chunk_size):
                    yield chunk

        for chunk in await asyncio.to_thread(read_file_chunks, audio_file, chunk_size):
            yield chunk

    async def tts(self, text: str, **kwargs) -> str:
        """
        Abstract async method for text-to-speech conversion.
        Must be implemented by subclasses.

        Args:
            text (str): The text to convert to speech
            **kwargs: Additional provider-specific parameters

        Returns:
            str: Path to the generated audio file

        Raises:
            NotImplementedError: If not implemented by subclass
        """
        raise NotImplementedError("Subclasses must implement the async tts method")



import json
import uuid
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream
from webscout.litagent import LitAgent


class HeckAI(Provider):
    """
    Provides an interface to interact with the HeckAI API using a LitAgent user-agent.

    This class supports conversational AI interactions with multiple available models,
    manages session state, handles streaming and non-streaming responses, and integrates
    with conversation history and prompt optimizers.

    Attributes:
        AVAILABLE_MODELS (list): List of supported model identifiers.
        url (str): API endpoint URL.
        session_id (str): Unique session identifier for the conversation.
        language (str): Language for the conversation.
        headers (dict): HTTP headers used for API requests.
        session (Session): curl_cffi session for HTTP requests.
        is_conversation (bool): Whether to maintain conversation history.
        max_tokens_to_sample (int): Maximum tokens to sample (not used by API).
        timeout (int): Request timeout in seconds.
        last_response (dict): Stores the last API response.
        model (str): Model identifier in use.
        previous_question (str): Last question sent to the API.
        previous_answer (str): Last answer received from the API.
        conversation (Conversation): Conversation history manager.
    """

    required_auth = False
    AVAILABLE_MODELS = [
        "google/gemini-2.5-flash-preview",
        "deepseek/deepseek-chat",
        "deepseek/deepseek-r1",
        "openai/gpt-4o-mini",
        "openai/gpt-4.1-mini",
        "x-ai/grok-3-mini-beta",
        "meta-llama/llama-4-scout",
        "openai/gpt-5-mini",
        "openai/gpt-5-nano",
    ]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 2049,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "google/gemini-2.5-flash-preview",
        language: str = "English",
    ):
        """
        Initializes the HeckAI API client.

        Args:
            is_conversation (bool): Whether to maintain conversation history.
            max_tokens (int): Maximum tokens to sample (not used by this API).
            timeout (int): Timeout for API requests in seconds.
            intro (str, optional): Introductory prompt for the conversation.
            filepath (str, optional): File path for storing conversation history.
            update_file (bool): Whether to update the conversation file.
            proxies (dict): Proxy settings for HTTP requests.
            history_offset (int): Offset for conversation history truncation.
            act (str, optional): Role or act for the conversation.
            model (str): Model identifier to use.
            language (str): Language for the conversation.

        Raises:
            ValueError: If the provided model is not in AVAILABLE_MODELS.
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.url = "https://api.heckai.weight-wave.com/api/ha/v1/chat"
        self.session_id = str(uuid.uuid4())
        self.language = language

        # Use LitAgent (keep if needed for other headers or logic)
        self.headers = {
            "Content-Type": "application/json",
            "Origin": "https://heck.ai",  # Keep Origin
            "Referer": "https://heck.ai/",  # Keep Referer
            "User-Agent": LitAgent().random(),  # Use random user agent
        }

        # Initialize curl_cffi Session
        self.session = Session()
        # Update curl_cffi session headers and proxies
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)  # Assign proxies directly

        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}
        self.model = model
        self.previous_question = None
        self.previous_answer = None

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        if act:
            self.conversation.intro = (
                AwesomePrompts().get_act(
                    cast(Union[str, int], act),
                    default=self.conversation.intro,
                    case_insensitive=True,
                )
                or self.conversation.intro
            )
        elif intro:
            self.conversation.intro = intro

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """
        Sends a prompt to the HeckAI API and returns the response.

        Args:
            prompt (str): The prompt or question to send to the API.
            stream (bool): If True, yields streaming responses as they arrive.
            raw (bool): If True, yields raw string chunks instead of dicts.
            optimizer (str, optional): Name of the optimizer to apply to the prompt.
            conversationally (bool): If True, optimizer is applied to the full conversation prompt.

        Returns:
            Union[Dict[str, Any], Generator]: If stream is False, returns a dict with the response text.
            If stream is True, yields response chunks as dicts or strings.

        Raises:
            Exception: If the optimizer is not available.
            exceptions.FailedToGenerateResponseError: On API or network errors.
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        # Payload construction
        payload = {
            "model": self.model,
            "question": conversation_prompt,
            "language": self.language,
            "sessionId": self.session_id,
            "previousQuestion": self.previous_question,
            "previousAnswer": self.previous_answer,
            "imgUrls": [],
            "superSmartMode": False,  # Added based on API request data
        }

        # Store this message as previous for next request
        self.previous_question = conversation_prompt

        def for_stream():
            streaming_text = ""  # Initialize outside try block
            try:
                response = self.session.post(
                    self.url,
                    data=json.dumps(payload),
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome110",
                )
                response.raise_for_status()

                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=1024),
                    intro_value="data: ",
                    to_json=False,
                    start_marker="data: [ANSWER_START]",
                    end_marker="data: [ANSWER_DONE]",
                    skip_markers=[
                        "data: [RELATE_Q_START]",
                        "data: [RELATE_Q_DONE]",
                        "data: [REASON_START]",
                        "data: [REASON_DONE]",
                    ],
                    yield_raw_on_error=True,
                    strip_chars="",
                    raw=raw,
                )

                for content_chunk in processed_stream:
                    if content_chunk and isinstance(content_chunk, str):
                        content_chunk = content_chunk.replace("\\\\", "\\").replace('\\"', '"')
                    if raw:
                        if content_chunk and isinstance(content_chunk, str):
                            streaming_text += content_chunk
                        yield content_chunk
                    else:
                        if content_chunk and isinstance(content_chunk, str):
                            streaming_text += content_chunk
                            yield dict(text=content_chunk)

                # Only update history if we received a valid response
                if streaming_text:
                    self.previous_answer = streaming_text
                    try:
                        if streaming_text and isinstance(streaming_text, str):
                            sanitized_text = streaming_text.strip()
                            if sanitized_text:
                                self.conversation.update_chat_history(prompt, sanitized_text)
                    except Exception as e:
                        print(f"Warning: Failed to update conversation history: {str(e)}")
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"Request failed (CurlError): {str(e)}"
                ) from e
            except Exception as e:
                err_text = ""
                if hasattr(e, "response"):
                    response_obj = getattr(e, "response")
                    if hasattr(response_obj, "text"):
                        err_text = getattr(response_obj, "text")
                raise exceptions.FailedToGenerateResponseError(
                    f"Request failed ({type(e).__name__}): {str(e)} - {err_text}"
                ) from e

        def for_non_stream():
            full_text = ""
            try:
                for chunk_data in for_stream():
                    if raw:
                        if isinstance(chunk_data, str):
                            chunk_data = chunk_data.replace("\\\\", "\\").replace('\\"', '"')
                            full_text += chunk_data
                    else:
                        if isinstance(chunk_data, dict) and "text" in chunk_data:
                            text = chunk_data["text"].replace("\\\\", "\\").replace('\\"', '"')
                            full_text += text
            except Exception as e:
                if not full_text:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Failed to get non-stream response: {str(e)}"
                    ) from e
            self.last_response = {"text": full_text}
            return full_text if raw else self.last_response

        return for_stream() if stream else for_non_stream()

    @staticmethod
    def fix_encoding(text):
        """
        Fixes encoding issues in the response text.

        Args:
            text (Union[str, dict]): The text or response dict to fix encoding for.

        Returns:
            Union[str, dict]: The text or dict with encoding corrected if possible.
        """
        if isinstance(text, dict) and "text" in text:
            try:
                val = str(text["text"]).encode("latin1").decode("utf-8")
                text["text"] = val.replace("\\\\", "\\").replace(
                    '\\"', '"'
                )  # Handle escaped backslashes
                return text
            except (UnicodeError, AttributeError):
                val = str(text["text"])
                text["text"] = val.replace("\\\\", "\\").replace(
                    '\\"', '"'
                )  # Handle escaped backslashes
                return text
        elif isinstance(text, str):
            try:
                return (
                    text.encode("latin1").decode("utf-8").replace("\\\\", "\\").replace('\\"', '"')
                )
            except (UnicodeError, AttributeError):
                return text.replace("\\\\", "\\").replace('\\"', '"')  # Handle escaped backslashes
        return text

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """
        Sends a prompt to the HeckAI API and returns only the message text.

        Args:
            prompt (str): The prompt or question to send to the API.
            stream (bool): If True, yields streaming response text.
            optimizer (str, optional): Name of the optimizer to apply to the prompt.
            conversationally (bool): If True, optimizer is applied to the full conversation prompt.
            **kwargs: Additional parameters including raw.

        Returns:
            Union[str, Generator[str, None, None]]: The response text, or a generator yielding text chunks.
        """
        raw = kwargs.get("raw", False)

        def for_stream_chat():
            # ask() yields dicts or strings when streaming
            gen = self.ask(
                prompt, stream=True, raw=raw, optimizer=optimizer, conversationally=conversationally
            )
            for response in gen:
                if raw:
                    yield cast(str, response)
                else:
                    yield self.get_message(cast(Response, response))

        def for_non_stream_chat():
            # ask() returns dict or str when not streaming
            response_data = self.ask(
                prompt,
                stream=False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            if raw:
                return response_data if isinstance(response_data, str) else str(response_data)
            return self.get_message(response_data)  # get_message expects dict

        return for_stream_chat() if stream else for_non_stream_chat()

    def get_message(self, response: Response) -> str:
        """
        Extracts the message text from the API response.

        Args:
            response (Response): The API response dictionary.

        Returns:
            str: The extracted message text. Returns an empty string if not found.
        """
        # Validate response format
        if not isinstance(response, dict):
            return str(response)

        # Handle missing text key gracefully
        if "text" not in response:
            return ""

        # Ensure text is a string
        resp_dict = cast(Dict[str, Any], response)
        text = cast(str, resp_dict["text"])
        if not isinstance(text, str):
            text = str(text)

        return text.replace("\\\\", "\\").replace('\\"', '"')


if __name__ == "__main__":
    from rich import print

    ai = HeckAI(model="openai/gpt-5-nano")
    response = ai.chat("tell me about humans", stream=True, raw=False)
    if hasattr(response, "__iter__") and not isinstance(response, (str, bytes)):
        for chunk in response:
            print(chunk, end="", flush=True)
    else:
        print(response)



import json
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union, cast

from curl_cffi.requests import Session

# Import LitAgent for user agent generation
# Import base classes and utility structures
from webscout.Provider.Openai_comp.base import (
    BaseChat,
    BaseCompletions,
    OpenAICompatibleProvider,
    SimpleModelList,
)
from webscout.Provider.Openai_comp.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
    count_tokens,
)


class Completions(BaseCompletions):
    def __init__(self, client: "LLMChat"):
        self._client = client

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = 2048,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        # In this case, we pass messages directly to the API
        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())

        if stream:
            return self._create_streaming(
                request_id, created_time, model, messages, max_tokens, timeout, proxies
            )
        else:
            return self._create_non_streaming(
                request_id, created_time, model, messages, max_tokens, timeout, proxies
            )

    def _create_streaming(
        self,
        request_id: str,
        created_time: int,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int],
        timeout: Optional[int],
        proxies: Optional[dict],
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            prompt_tokens = count_tokens(json.dumps(messages))
            completion_tokens = 0

            url = f"{self._client.api_endpoint}?model={model}"
            payload = {"messages": messages, "max_tokens": max_tokens or 2048, "stream": True}

            response = self._client.session.post(
                url,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                impersonate="chrome110",
            )
            response.raise_for_status()

            full_content = ""
            for line in response.iter_lines():
                if line:
                    line = line.decode("utf-8")
                    if line.startswith("data: "):
                        data_str = line[6:]
                        if data_str.strip() == "[DONE]":
                            break

                        try:
                            data = json.loads(data_str)
                            content = data.get("response", "")
                            if content:
                                full_content += content
                                completion_tokens += 1

                                delta = ChoiceDelta(content=content, role="assistant")
                                choice = Choice(index=0, delta=delta, finish_reason=None)
                                chunk = ChatCompletionChunk(
                                    id=request_id,
                                    choices=[choice],
                                    created=created_time,
                                    model=model,
                                )
                                yield chunk
                        except json.JSONDecodeError:
                            continue

            # Final chunk
            delta = ChoiceDelta(content=None)
            choice = Choice(index=0, delta=delta, finish_reason="stop")
            final_chunk = ChatCompletionChunk(
                id=request_id, choices=[choice], created=created_time, model=model
            )
            usage_obj = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=count_tokens(full_content),
                total_tokens=prompt_tokens + count_tokens(full_content),
            )
            final_chunk.usage = usage_obj.model_dump(exclude_none=True)
            yield final_chunk

        except Exception as e:
            raise IOError(f"LLMChat streaming request failed: {e}") from e

    def _create_non_streaming(
        self,
        request_id: str,
        created_time: int,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int],
        timeout: Optional[int],
        proxies: Optional[dict],
    ) -> ChatCompletion:
        try:
            full_content = ""
            prompt_tokens = count_tokens(json.dumps(messages))

            for chunk in self._create_streaming(
                request_id, created_time, model, messages, max_tokens, timeout, proxies
            ):
                if chunk.choices[0].delta and chunk.choices[0].delta.content:
                    full_content += chunk.choices[0].delta.content

            message = ChatCompletionMessage(role="assistant", content=full_content)
            choice = Choice(index=0, message=message, finish_reason="stop")
            usage = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=count_tokens(full_content),
                total_tokens=prompt_tokens + count_tokens(full_content),
            )

            return ChatCompletion(
                id=request_id, choices=[choice], created=created_time, model=model, usage=usage
            )
        except Exception as e:
            raise IOError(f"LLMChat request failed: {e}") from e


class Chat(BaseChat):
    def __init__(self, client: "LLMChat"):
        self.completions = Completions(client)


class LLMChat(OpenAICompatibleProvider):
    required_auth = False
    AVAILABLE_MODELS = [
        "@cf/aisingapore/gemma-sea-lion-v4-27b-it",
        "@cf/deepseek-ai/deepseek-math-7b-instruct",
        "@cf/deepseek-ai/deepseek-r1-distill-qwen-32b",
        "@cf/defog/sqlcoder-7b-2",
        "@cf/fblgit/una-cybertron-7b-v2-bf16",
        "@cf/google/gemma-2b-it-lora",
        "@cf/google/gemma-3-12b-it",
        "@cf/ibm-granite/granite-4.0-h-micro",
        "@cf/meta-llama/llama-2-7b-chat-hf-lora",
        "@cf/meta/llama-2-7b-chat-fp16",
        "@cf/meta/llama-2-7b-chat-int8",
        "@cf/meta/llama-3-8b-instruct",
        "@cf/meta/llama-3-8b-instruct-awq",
        "@cf/meta/llama-3.1-70b-instruct",
        "@cf/meta/llama-3.1-8b-instruct",
        "@cf/meta/llama-3.2-1b-instruct",
        "@cf/meta/llama-3.2-3b-instruct",
        "@cf/meta/llama-3.3-70b-instruct-fp8-fast",
        "@cf/meta/llama-4-scout-17b-16e-instruct",
        "@cf/meta/llama/llama-2-7b-chat-hf-lora",
        "@cf/meta/meta-llama-3-8b-instruct",
        "@cf/microsoft/phi-2",
        "@cf/mistral/mistral-7b-instruct-v0.1-vllm",
        "@cf/mistral/mistral-7b-instruct-v0.2-lora",
        "@cf/mistralai/mistral-small-3.1-24b-instruct",
        "@cf/openchat/openchat-3.5-0106",
        "@cf/qwen/qwen1.5-0.5b-chat",
        "@cf/qwen/qwen1.5-1.8b-chat",
        "@cf/qwen/qwen1.5-14b-chat-awq",
        "@cf/qwen/qwen1.5-7b-chat-awq",
        "@cf/qwen/qwen2.5-coder-32b-instruct",
        "@cf/qwen/qwen3-30b-a3b-fp8",
        "@cf/qwen/qwq-32b",
        "@cf/tiiuae/falcon-7b-instruct",
        "@cf/tinyllama/tinyllama-1.1b-chat-v1.0",
        "@hf/google/gemma-7b-it",
        "@hf/meta-llama/meta-llama-3-8b-instruct",
        "@hf/mistral/mistral-7b-instruct-v0.2",
        "@hf/nexusflow/starling-lm-7b-beta",
        "@hf/thebloke/deepseek-coder-6.7b-base-awq",
        "@hf/thebloke/deepseek-coder-6.7b-instruct-awq",
        "@hf/thebloke/llama-2-13b-chat-awq",
        "@hf/thebloke/llamaguard-7b-awq",
        "@hf/thebloke/mistral-7b-instruct-v0.1-awq",
        "@hf/thebloke/neural-chat-7b-v3-1-awq",
        "@hf/thebloke/openhermes-2.5-mistral-7b-awq",
        "@hf/thebloke/zephyr-7b-beta-awq",
    ]

    def __init__(self, proxies: dict = {}, timeout: int = 30):
        self.session = Session()
        self.timeout = timeout
        self.api_endpoint = "https://llmchat.in/inference/stream"
        self.proxies = proxies
        if proxies:
            self.session.proxies.update(cast(Any, proxies))

        self.headers = {
            "Content-Type": "application/json",
            "Accept": "*/*",
            "Origin": "https://llmchat.in",
            "Referer": "https://llmchat.in/",
        }
        self.session.headers.update(self.headers)
        self.chat = Chat(self)

    @property
    def models(self) -> SimpleModelList:
        return SimpleModelList(type(self).AVAILABLE_MODELS)


if __name__ == "__main__":
    client = LLMChat()
    response = client.chat.completions.create(
        model="@cf/meta/llama-3.1-70b-instruct",
        messages=[{"role": "user", "content": "Say 'Hello' in one word"}],
        stream=True,
    )
    for chunk in response:
        if hasattr(chunk, "choices") and chunk.choices:
            choices = cast(Any, chunk.choices)
            delta = choices[0].delta
            if delta and hasattr(delta, "content") and delta.content:
                print(delta.content, end="", flush=True)
